#ifndef _BITS_HH
#define _BITS_HH 1

#include <stdint.h>

#include "util/array.hh"
#include "util/misc.hh"

#include "bitcount.hh"
#include "bits_base.hh"

#define B16

namespace {
  inline uint8_t bitcount(uint8_t x) {
    return __bitcount8[x];
  }

  inline uint8_t bitcount(uint16_t x) {
#ifdef B16
    return __bitcount16[x];
#else
    return __bitcount8[x >> 0 & 0xff] + __bitcount8[x >> 8 & 0xff];
#endif
  }

  inline uint8_t bitcount(uint32_t x) {
#ifdef B16
    return __bitcount16[x >> 0 & 0xffff] + __bitcount16[x >> 16 & 0xffff];
#else
    return __bitcount8[x >> 0 & 0xff] + __bitcount8[x >> 8 & 0xff] +
      __bitcount8[x >> 16 & 0xff] + __bitcount8[x >> 24 & 0xff];
#endif
  }

  inline uint8_t bitcount(uint64_t x) {
#ifdef B16
    return __bitcount16[x >> 0 & 0xffff] + __bitcount16[x >> 16 & 0xffff]
      + __bitcount16[x >> 32 & 0xffff] + __bitcount16[x >> 48 & 0xffff];
#else
    return __bitcount8[x >> 0 & 0xff] + __bitcount8[x >> 8 & 0xff] +
      __bitcount8[x >> 16 & 0xff] + __bitcount8[x >> 24 & 0xff] +
      __bitcount8[x >> 32 & 0xff] + __bitcount8[x >> 40 & 0xff] +
      __bitcount8[x >> 48 & 0xff] + __bitcount8[x >> 56 & 0xff];
#endif
  }
}

template <typename Base>
class BitArray {
 public:
  typedef BitArraySiblings<Base> SiblingTypes;

  BitArray() : size(0) {};
  BitArray(size_t n);

  bool Get(size_t n) const;
  void Set(size_t n);
  void Reset(size_t n);

  bool operator[](size_t n) const;

  const static size_t BaseSize = sizeof(Base) * 8;

  size_t size;
  SArray<Base> bits;
};

template <typename Base>
BitArray<Base>::BitArray(size_t n) :
  size(n),
  bits(ceildiv(n, BaseSize), static_cast<Base>(0))
{
}

template <typename Base>
bool BitArray<Base>::Get(size_t n) const {
  return (bits[n / BaseSize] >> (n % BaseSize)) & 1;
}

template <typename Base>
void BitArray<Base>::Set(size_t n) {
  bits[n / BaseSize] |= (1ul << (n % BaseSize));
}

template <typename Base>
void BitArray<Base>::Reset(size_t n) {
  bits[n / BaseSize] &= ~(1ul << (n % BaseSize));
}

template <typename Base>
bool BitArray<Base>::operator[](size_t n) const {
  return Get(n);
}

template <typename Base, typename RankT, size_t Rate>
class BitRankSL : public BitArray<Base> {
 public:
  typedef BitRankSiblings<Base, RankT, Rate> SiblingTypes;

  BitRankSL() {};
  BitRankSL(RankT n);

  void CalculateRanks();
  RankT Rank(RankT t) const;
  RankT Rank0(RankT t) const;
  RankT RankM(RankT t, int zero) const;

  using BitArray<Base>::size;
  SArray<RankT> rs;

  using BitArray<Base>::bits;
  using BitArray<Base>::BaseSize;
  static const size_t s = BaseSize * Rate;

 private:

  friend class boost::serialization::access;
  template<class Archive>
  void serialize(Archive & ar, const unsigned int version);
};

template <typename Base, typename RankT, size_t Rate>
template<class Archive>
void BitRankSL<Base, RankT, Rate>::serialize(Archive & ar, const unsigned int version) {
  ar & size;
  ar & bits;
  ar & rs;
}


template <typename Base, typename RankT, size_t Rate>
BitRankSL<Base, RankT, Rate>::BitRankSL(RankT n) :
  // n + 1 - see Rank()
  BitArray<Base>(n + 1),
  rs(ceildiv(n + 1, BaseSize * Rate))
{
  // For users the size should be n
  size = n;
}

template <typename Base, typename RankT, size_t Rate>
void BitRankSL<Base, RankT, Rate>::CalculateRanks() {
  //std::cerr << "base: " << BaseSize << " rate: " << Rate << " -> " << rs.size << "\n";
  rs[0] = 0;
  for (size_t i = 1 ; i < rs.size ; ++i) {
    rs[i] = rs[i - 1];
    for (size_t r = 0 ; r < Rate ; ++r) {
      rs[i] += bitcount(bits[(i - 1) * Rate + r]);
      //std::cerr << "bits[" << (i - 1) * Rate + r << "] = " << (int)bitcount(bits[(i - 1) * Rate + r]) << "\n";
    }
    //std::cerr << "rs[" << i << "] = " << rs[i] << "\n";
  }
}

template <typename Base, typename RankT, size_t Rate>
inline RankT BitRankSL<Base, RankT, Rate>::Rank(RankT n) const {
  ++n;
  RankT r = rs[n / s];
  //std::cerr << "rs[" << n / s << "] = " << rs[n / s] << "\n";
  for (size_t i = (n / s) * Rate ; i < n / BaseSize ; ++i) {
    r += bitcount(bits[i]);
    //std::cerr << "bits[" << i << "] = " << (int)bitcount(bits[i]) << "\n";
  }
  // first n % BaseSize bits from n / BaseSize
  //r += bitcount(static_cast<Base>(bits[n / BaseSize] & static_cast<Base>((1ull << (n % BaseSize)) - 1)));
  //if (n % BaseSize)
  //  r += bitcount(static_cast<Base>(bits[n / BaseSize] << (BaseSize - n % BaseSize)));
  r += bitcount(static_cast<Base>(bits[n / BaseSize] & static_cast<Base>(one_masks[n % BaseSize])));

  return r;
}

template <typename Base, typename RankT, size_t Rate>
inline RankT BitRankSL<Base, RankT, Rate>::Rank0(RankT n) const {
  return n - Rank(n) + 1;
}

template <typename Base, typename RankT, size_t Rate>
inline RankT BitRankSL<Base, RankT, Rate>::RankM(RankT n, int ones) const {
  RankT r = Rank(n);
#if 0
  r = (ones ^ (ones - 1)) * r + (ones ^ 1) * (n + 1);
#else
  if (ones == 0)
    r = n - r + 1;
#endif
  return r;
}

template <typename Base>
std::ostream & operator<<(std::ostream & o, const BitArray<Base> & ba) {
  for (size_t i = 0 ; i < ba.bits.size ; ++i) {
    for (size_t j = 0 ; j < BitArray<Base>::BaseSize ; ++j)
      o << (ba[i * BitArray<Base>::BaseSize + j] ? "1" : "0");
    o << " ";
  }

  return o;
}

#endif
