#ifndef _BITS_CUDA_HH
#define _BITS_CUDA_HH 1

#include <assert.h>

#include <cuda_runtime.h>

#include "util/array_cuda.hh"
#include "util/misc.hh"

#include "bitcount.hh"
#include "bits_base.hh"
#include "bits.hh"

//#define USE_UNION
extern __device__ int __popc(unsigned int);
extern __device__ int __popcll(unsigned long long int);

namespace cuda {
  __device__ inline uint8_t bitcount(uint32_t x) {
    return __popc(x);
  }

  __device__ inline uint8_t bitcount(uint64_t x) {
    return __popcll(x);
  }
}

template <typename Base>
class KernelBitArray {
 public:
  typedef BitArraySiblings<Base> SiblingTypes;

  __device__ bool Get(size_t n) const;

  __device__ bool operator[](size_t n) const;

  const static size_t BaseSize = sizeof(Base) * 8;
  size_t size;

  KernelArray<Base> bits;
};

template <typename Base>
__device__ bool KernelBitArray<Base>::Get(size_t n) const {
  return (bits[n / BaseSize] >> (n % BaseSize)) & 1;
}

template <typename Base>
__device__ bool KernelBitArray<Base>::operator[](size_t n) const {
  return Get(n);
}

template <typename Base>
class CudaBitArray {
  typedef BitArraySiblings<Base> SiblingTypes;
  typedef typename SiblingTypes::Kernel KernelBA;

 public:
  CudaBitArray(const BitArray<Base> & ba);
  CudaBitArray() { }
 
  KernelBA ForKernel() const { KernelBA ka; ka.bits = bits.ForKernel(); ka.size = size; return ka; }

 protected:  
  size_t size;
  CudaArray<Base> bits;
 
};


template <typename Base>
CudaBitArray<Base>::CudaBitArray(const BitArray<Base> & ba) :
  size(ba.size),
  bits(ba.bits)
{
}

template <typename Base, typename RankT, size_t Rate>
class KernelBitRankSL : public KernelBitArray<Base> {
public:
  typedef BitRankSiblings<Base, RankT, Rate> SiblingTypes;

  __device__ RankT Rank(RankT t) const;
  __device__ RankT Rank0(RankT t) const;
  __device__ RankT RankM(RankT t, int zero) const;

  using KernelBitArray<Base>::bits;
  using KernelBitArray<Base>::BaseSize;
  using KernelBitArray<Base>::size;
  static const size_t s = BaseSize * Rate;

  KernelArray<RankT> rs;

};

template <typename Base, typename RankT, size_t Rate>
class CudaBitRankSL : public CudaBitArray<Base> {
  typedef BitRankSiblings<Base, RankT, Rate> SiblingTypes;
  typedef typename SiblingTypes::Kernel KernelBR;

 public:
  CudaBitRankSL() { }
  CudaBitRankSL(const BitRankSL<Base, RankT, Rate> & br);

  using CudaBitArray<Base>::size;
  using CudaBitArray<Base>::bits;

  KernelBR ForKernel() const;

 private:
  CudaArray<RankT> rs;

};

template <typename Base, typename RankT, size_t Rate>
CudaBitRankSL<Base, RankT, Rate>::CudaBitRankSL(const BitRankSL<Base, RankT, Rate> & br) :
  CudaBitArray<Base>(br),
  rs(br.rs)
{
}

template <typename Base, typename RankT, size_t Rate>
typename CudaBitRankSL<Base, RankT, Rate>::KernelBR CudaBitRankSL<Base, RankT, Rate>::ForKernel() const
{
  KernelBR ka;
  ka.rs = rs.ForKernel();
  ka.bits = bits.ForKernel();
  ka.size = size;
  return ka;
}


template <typename Base, typename RankT, size_t Rate>
__device__ inline RankT KernelBitRankSL<Base, RankT, Rate>::Rank(RankT n) const {
  ++n;
  RankT r = rs[n / s];
  //std::cerr << "rs[" << n / s << "] = " << rs[n / s] << "\n";
  for (size_t i = (n / s) * Rate ; i < n / BaseSize ; ++i) {
    r += cuda::bitcount(bits[i]);
    //std::cerr << "bits[" << i << "] = " << (int)cuda::bitcount(bits[i]) << "\n";
  }
  // first n % BaseSize bits from n / BaseSize
  //r += cuda::bitcount(static_cast<Base>(bits[n / BaseSize] & static_cast<Base>((1ull << (n % BaseSize)) - 1)));
  //if (n % BaseSize)
  
  r += cuda::bitcount(static_cast<Base>(bits[n / BaseSize] << (BaseSize - n % BaseSize)));

  return r;
}

template <typename Base, typename RankT, size_t Rate>
__device__ inline RankT KernelBitRankSL<Base, RankT, Rate>::Rank0(RankT n) const {
  return n - Rank(n) + 1;
}

template <typename Base, typename RankT, size_t Rate>
__device__ inline RankT KernelBitRankSL<Base, RankT, Rate>::RankM(RankT n, int ones) const {
  RankT r = Rank(n);
#if 0
  r = (ones ^ (ones - 1)) * r + (ones ^ 1) * (n + 1);
#else
  if (ones == 0)
    r = n - r + 1;
#endif
  return r;
}


template <typename Base = uint32_t, typename RankT = uint32_t, size_t Rate = 3>
struct Package {
/*  Base data[Rate + 1];

  __host__ __device__ __forceinline__ Base operator[](RankT n) const { return data[n]; }
  __host__ __device__ __forceinline__ Base & operator[](RankT n) { return data[n]; }
*/
#ifdef USE_UNION
  union {
    ulonglong2 data;
    uint32_t u32[4];
  } data;
#else
  ulonglong2 data;
#endif

  __host__ __device__ Base operator[](RankT n) const {
#ifdef USE_UNION
    return data.u32[n];
#else
    uint64_t d = n / 2 ? data.y : data.x;
    return (d >> ((n & 1) * 32)) & 0xffffffff;
#endif
 }

  __host__ void Set(RankT n, uint64_t d) {
#ifdef USE_UNION
    data.u32[n] = d;
#else
    const int offset = ((n & 1) * 32);
    const uint64_t mask = 0xffffffff00000000 >> offset;
    if (n / 2) {
      data.y &= mask;
      data.y |= d << offset;
    } else {
      data.x &= mask;
      data.x |= d << offset;
    }
#endif
  }
  //__host__ __device__ Base & operator[](RankT n) { return data.u32[n]; }
};
#undef USE_UNION

template <typename Base, typename RankT, size_t Rate>
class KernelBitRankPacked {
public:
  typedef BitRankSiblings<Base, RankT, Rate> SiblingTypes;

  __device__ bool Get(size_t n) const;
  __device__ bool operator[](size_t n) const;
  __device__ RankT Rank(RankT t) const;
  __device__ RankT Rank0(RankT t) const;
  __device__ RankT RankM(RankT t, int zero) const;

  const static size_t BaseSize = sizeof(Base) * 8;
  static const size_t BaseRateSize = BaseSize * Rate;

  size_t size;
  KernelArray<Package<Base, RankT, Rate> > data;
};
 
template <typename Base, typename RankT, size_t Rate>
__device__ bool KernelBitRankPacked<Base, RankT, Rate>::Get(size_t n) const
{
  size_t i = n / BaseRateSize;
  size_t j = n % BaseRateSize;
  return (data[i][1 + j / BaseSize] >> (j % BaseSize)) & 1;
}

template <typename Base, typename RankT, size_t Rate>
__device__ bool KernelBitRankPacked<Base, RankT, Rate>::operator[](size_t n) const
{
  return Get(n);
}

template <typename Base, typename RankT, size_t Rate>
__device__ inline RankT KernelBitRankPacked<Base, RankT, Rate>::Rank(RankT n) const
{
  ++n;

  Package<Base, RankT, Rate> package = data[n / BaseRateSize];
  RankT bit = n % BaseRateSize;
  RankT chunk = bit / BaseSize;

  RankT r = package[0];

  for (RankT i = 1; i < 1 + chunk; i++) {
    r += cuda::bitcount(package[i]);
  }

  r += cuda::bitcount(static_cast<Base>(package[1 + chunk] << (BaseSize - bit % BaseSize)));
  return r;
}

template <typename Base, typename RankT, size_t Rate>
__device__ inline RankT KernelBitRankPacked<Base, RankT, Rate>::Rank0(RankT n) const
{
  return n - Rank(n) + 1;
}

template <typename Base, typename RankT, size_t Rate>
__device__ inline RankT KernelBitRankPacked<Base, RankT, Rate>::RankM(RankT n, int ones) const
{
  RankT r = Rank(n);
#if 1
  r = (ones ^ (ones - 1)) * r + (ones ^ 1) * (n + 1);
#else
  if (ones == 0)
    r = n - r + 1;
#endif
  return r;
}


template <typename Base, typename RankT, size_t Rate>
class CudaBitRankPacked {
  typedef BitRankSiblings<Base, RankT, Rate> SiblingTypes;
  typedef typename SiblingTypes::KernelPacked KernelBRP;

 public:
  CudaBitRankPacked() { }
  CudaBitRankPacked(const BitRankSL<Base, RankT, Rate> & br);

  KernelBRP ForKernel() const { KernelBRP ka; ka.data = data.ForKernel(); ka.size = size; return ka; }

 protected:
  size_t size, data_size;
  CudaArray<Package<Base, RankT, Rate> > data;
};

template <typename Base, typename RankT, size_t Rate>
CudaBitRankPacked<Base, RankT, Rate>::CudaBitRankPacked(const BitRankSL<Base, RankT, Rate> & br) :
  size(br.size),
  data_size(br.rs.size)
{
  SArray<Package<Base, RankT, Rate> > tmp(data_size);
  
  for(RankT i = 0; i < data_size; i++)
  {
    tmp[i].Set(0, br.rs[i]);
    for(RankT j = 0; j < Rate; j++)
    {
      size_t k = i * Rate + j;
      if(k < br.bits.size)
        tmp[i].Set(1 + j, br.bits[k]);
    }
  }

  data = CudaArray<Package<Base, RankT, Rate> >(tmp);
}


#endif
