#ifndef _INDEX_HH
#define _INDEX_HH

#define BOOST_MPL_CFG_NO_PREPROCESSED_HEADERS
#define BOOST_MPL_LIMIT_SET_SIZE 50
#define BOOST_MPL_LIMIT_STRING_SIZE 64
#include <boost/mpl/set.hpp>
#include <boost/mpl/map.hpp>
#include <boost/mpl/int.hpp>
#include <boost/mpl/at.hpp>
#include <boost/mpl/vector.hpp>
#include <boost/mpl/string.hpp>
#include <boost/mpl/filter_view.hpp>

#include "config.hh"
#include "util/static_for_each.hh"
#include "util/siblings.hh"
#include "ssa.hh"
#include "dna.hh"

namespace bmap {

struct CharMeta {
  typedef unsigned char char_type;
  typedef uint32_t size_type;
  typedef SArray<size_type> size_array;
  static const int bits_per_char = 8;
  static const char_type unknown_char = 0;
  typedef SArray<char_type> string_type;
  typedef CompactArray<unsigned long, bits_per_char, char_type> compact_type;
};

// types and consts used by the index
template <typename bitv_b, int bitv_f, int srb=0, int srt=0>
struct DnaMeta {
  typedef dna_t char_type;
  typedef uint32_t size_type;
  typedef SArray<size_type> size_array;
  typedef CSibling<size_array>::Type cuda_size_array;
  typedef KSibling<size_array>::Type kernel_size_array;
  typedef bitv_b bitv_base;

  static const int bits_per_char = 2;
  static const int sigma_size = 4;
  static const char_type unknown_char = dna::N;

  static const int state_size = 128; // max query lenght for some Indexes

  static const int spep_cache_size = 10;

  static const int sample_rate_text = srt;
  static const int sample_rate_bwt = srb;
  static const int bitv_factor = bitv_f;

  typedef DnaString string_type;
  typedef CSibling<string_type>::Type cuda_string_type;
  typedef Siblings<string_type>::Types::KernelSized kernel_string_type;
  typedef DnaStrings strings_type;
  typedef CSibling<strings_type>::Type cuda_strings_type;
  typedef KSibling<strings_type>::Type kernel_strings_type;
  typedef DnaCompact compact_type;
  typedef Siblings<compact_type>::Types::Cuda cuda_compact_type;
  typedef Siblings<compact_type>::Types::Kernel kernel_compact_type;
  typedef BitRankSL<bitv_base, uint32_t, bitv_factor> wt_rank_type;
  typedef typename Siblings<wt_rank_type>::Types::CudaPacked cuda_wt_rank_type;
  typedef typename Siblings<wt_rank_type>::Types::KernelPacked kernel_wt_rank_type;
  typedef BitRankSL<bitv_base, uint32_t, bitv_factor> sampled_bwt_rank_type;
  typedef typename Siblings<sampled_bwt_rank_type>::Types::CudaPacked cuda_sampled_bwt_rank_type;
  typedef typename Siblings<sampled_bwt_rank_type>::Types::KernelPacked kernel_sampled_bwt_rank_type;
};

// Signature of an index
struct IndexSig {
  std::string index_name;

  IndexSig() {}
  IndexSig(const std::string & name) : index_name(name) {}

  bool operator==(const IndexSig & other) const {
    return index_name == other.index_name;
  }

  friend class boost::serialization::access;
  template <typename Archive>
  void serialize(Archive & ar, const unsigned int version) {
    ar & index_name;
  }
};

template <typename IndexName, typename IndexType, typename HostSupport, typename CudaSupport>
struct IndexInfo : boost::mpl::vector<IndexName, IndexType, HostSupport, CudaSupport> {
  typedef IndexName name;
  typedef boost::mpl::c_str<name> name_c_str;
  typedef IndexType index_type;
  typedef HostSupport host_support;
  typedef CudaSupport cuda_support;
};

struct IndexMeta {
  typedef boost::mpl::set<
    IndexInfo<boost::mpl::string<'s','s','a','_','6','4','_','1','_','8','_','3','2'>, SSA<DnaMeta<uint64_t, 1, 8, 32> >, boost::mpl::true_, boost::mpl::false_>,
    IndexInfo<boost::mpl::string<'s','s','a','_','6','4','_','2','_','8','_','3','2'>, SSA<DnaMeta<uint64_t, 2, 8, 32> >, boost::mpl::true_, boost::mpl::false_>,
    IndexInfo<boost::mpl::string<'s','s','a','_','6','4','_','4','_','8','_','3','2'>, SSA<DnaMeta<uint64_t, 4, 8, 32> >, boost::mpl::true_, boost::mpl::false_>,
    IndexInfo<boost::mpl::string<'s','s','a','_','6','4','_','1','_','8','_','6','4'>, SSA<DnaMeta<uint64_t, 1, 8, 64> >, boost::mpl::true_, boost::mpl::false_>,
    IndexInfo<boost::mpl::string<'s','s','a','_','6','4','_','2','_','8','_','6','4'>, SSA<DnaMeta<uint64_t, 2, 8, 64> >, boost::mpl::true_, boost::mpl::false_>,
    IndexInfo<boost::mpl::string<'s','s','a','_','6','4','_','4','_','8','_','6','4'>, SSA<DnaMeta<uint64_t, 4, 8, 64> >, boost::mpl::true_, boost::mpl::false_>,
    IndexInfo<boost::mpl::string<'s','s','a','_','6','4','_','1','_','1','6','_','3','2'>, SSA<DnaMeta<uint64_t, 1, 16, 32> >, boost::mpl::true_, boost::mpl::false_>,
    IndexInfo<boost::mpl::string<'s','s','a','_','6','4','_','2','_','1','6','_','3','2'>, SSA<DnaMeta<uint64_t, 2, 16, 32> >, boost::mpl::true_, boost::mpl::false_>,
    IndexInfo<boost::mpl::string<'s','s','a','_','6','4','_','4','_','1','6','_','3','2'>, SSA<DnaMeta<uint64_t, 4, 16, 32> >, boost::mpl::true_, boost::mpl::false_>,
    IndexInfo<boost::mpl::string<'s','s','a','_','6','4','_','1','_','1','6','_','6','4'>, SSA<DnaMeta<uint64_t, 1, 16, 64> >, boost::mpl::true_, boost::mpl::false_>,
    IndexInfo<boost::mpl::string<'s','s','a','_','6','4','_','2','_','1','6','_','6','4'>, SSA<DnaMeta<uint64_t, 2, 16, 64> >, boost::mpl::true_, boost::mpl::false_>,
    IndexInfo<boost::mpl::string<'s','s','a','_','6','4','_','4','_','1','6','_','6','4'>, SSA<DnaMeta<uint64_t, 4, 16, 64> >, boost::mpl::true_, boost::mpl::false_>,
    IndexInfo<boost::mpl::string<'s','s','a','_','6','4','_','1','_','3','2','_','3','2'>, SSA<DnaMeta<uint64_t, 1, 32, 32> >, boost::mpl::true_, boost::mpl::false_>,
    IndexInfo<boost::mpl::string<'s','s','a','_','6','4','_','2','_','3','2','_','3','2'>, SSA<DnaMeta<uint64_t, 2, 32, 32> >, boost::mpl::true_, boost::mpl::false_>,
    IndexInfo<boost::mpl::string<'s','s','a','_','6','4','_','4','_','3','2','_','3','2'>, SSA<DnaMeta<uint64_t, 4, 32, 32> >, boost::mpl::true_, boost::mpl::false_>,
    IndexInfo<boost::mpl::string<'s','s','a','_','6','4','_','1','_','3','2','_','6','4'>, SSA<DnaMeta<uint64_t, 1, 32, 64> >, boost::mpl::true_, boost::mpl::false_>,
    IndexInfo<boost::mpl::string<'s','s','a','_','6','4','_','2','_','3','2','_','6','4'>, SSA<DnaMeta<uint64_t, 2, 32, 64> >, boost::mpl::true_, boost::mpl::false_>,
    IndexInfo<boost::mpl::string<'s','s','a','_','6','4','_','4','_','3','2','_','6','4'>, SSA<DnaMeta<uint64_t, 4, 32, 64> >, boost::mpl::true_, boost::mpl::false_>,
    IndexInfo<boost::mpl::string<'s','s','a','_','6','4','_','1','_','6','4','_','3','2'>, SSA<DnaMeta<uint64_t, 1, 64, 32> >, boost::mpl::true_, boost::mpl::false_>,
    IndexInfo<boost::mpl::string<'s','s','a','_','6','4','_','2','_','6','4','_','3','2'>, SSA<DnaMeta<uint64_t, 2, 64, 32> >, boost::mpl::true_, boost::mpl::false_>,
    IndexInfo<boost::mpl::string<'s','s','a','_','6','4','_','4','_','6','4','_','3','2'>, SSA<DnaMeta<uint64_t, 4, 64, 32> >, boost::mpl::true_, boost::mpl::false_>,
    IndexInfo<boost::mpl::string<'s','s','a','_','6','4','_','1','_','6','4','_','6','4'>, SSA<DnaMeta<uint64_t, 1, 64, 64> >, boost::mpl::true_, boost::mpl::false_>,
    IndexInfo<boost::mpl::string<'s','s','a','_','6','4','_','2','_','6','4','_','6','4'>, SSA<DnaMeta<uint64_t, 2, 64, 64> >, boost::mpl::true_, boost::mpl::false_>,
    IndexInfo<boost::mpl::string<'s','s','a','_','6','4','_','4','_','6','4','_','6','4'>, SSA<DnaMeta<uint64_t, 4, 64, 64> >, boost::mpl::true_, boost::mpl::false_>,
    IndexInfo<boost::mpl::string<'s','s','a','t','_','6','4','_','1','_','8'>, SSAT<DnaMeta<uint64_t, 1, 8> >, boost::mpl::true_, boost::mpl::false_>,
    IndexInfo<boost::mpl::string<'s','s','a','t','_','6','4','_','2','_','8'>, SSAT<DnaMeta<uint64_t, 2, 8> >, boost::mpl::true_, boost::mpl::false_>,
    IndexInfo<boost::mpl::string<'s','s','a','t','_','6','4','_','4','_','8'>, SSAT<DnaMeta<uint64_t, 4, 8> >, boost::mpl::true_, boost::mpl::false_>,
    IndexInfo<boost::mpl::string<'s','s','a','t','_','6','4','_','1','_','1','6'>, SSAT<DnaMeta<uint64_t, 1, 16> >, boost::mpl::true_, boost::mpl::false_>,
    IndexInfo<boost::mpl::string<'s','s','a','t','_','6','4','_','2','_','1','6'>, SSAT<DnaMeta<uint64_t, 2, 16> >, boost::mpl::true_, boost::mpl::false_>,
    IndexInfo<boost::mpl::string<'s','s','a','t','_','6','4','_','4','_','1','6'>, SSAT<DnaMeta<uint64_t, 4, 16> >, boost::mpl::true_, boost::mpl::false_>,
    IndexInfo<boost::mpl::string<'s','s','a','t','_','6','4','_','1','_','3','2'>, SSAT<DnaMeta<uint64_t, 1, 32> >, boost::mpl::true_, boost::mpl::false_>,
    IndexInfo<boost::mpl::string<'s','s','a','t','_','6','4','_','2','_','3','2'>, SSAT<DnaMeta<uint64_t, 2, 32> >, boost::mpl::true_, boost::mpl::false_>,
    IndexInfo<boost::mpl::string<'s','s','a','t','_','6','4','_','4','_','3','2'>, SSAT<DnaMeta<uint64_t, 4, 32> >, boost::mpl::true_, boost::mpl::false_>,
    IndexInfo<boost::mpl::string<'s','s','a','t','_','6','4','_','1','_','6','4'>, SSAT<DnaMeta<uint64_t, 1, 64> >, boost::mpl::true_, boost::mpl::false_>,
    IndexInfo<boost::mpl::string<'s','s','a','t','_','6','4','_','2','_','6','4'>, SSAT<DnaMeta<uint64_t, 2, 64> >, boost::mpl::true_, boost::mpl::false_>,
    IndexInfo<boost::mpl::string<'s','s','a','t','_','6','4','_','4','_','6','4'>, SSAT<DnaMeta<uint64_t, 4, 64> >, boost::mpl::true_, boost::mpl::false_>,
    IndexInfo<boost::mpl::string<'c','s','s','a','t','_','3','2','_','3','_','8'>, SSAT<DnaMeta<uint32_t, 3, 8> >, boost::mpl::true_, boost::mpl::true_>,
    IndexInfo<boost::mpl::string<'c','s','s','a','t','_','3','2','_','3','_','1','6'>, SSAT<DnaMeta<uint32_t, 3, 16> >, boost::mpl::true_, boost::mpl::true_>,
    IndexInfo<boost::mpl::string<'c','s','s','a','t','_','3','2','_','3','_','3','2'>, SSAT<DnaMeta<uint32_t, 3, 32> >, boost::mpl::true_, boost::mpl::true_>,
    IndexInfo<boost::mpl::string<'c','s','s','a','t','_','3','2','_','3','_','6','4'>, SSAT<DnaMeta<uint32_t, 3, 64> >, boost::mpl::true_, boost::mpl::true_>,
    IndexInfo<boost::mpl::string<'s','s','a','t','s','_','6','4','_','1'>, SSATS<DnaMeta<uint64_t, 1> >, boost::mpl::true_, boost::mpl::false_>,
    IndexInfo<boost::mpl::string<'s','s','a','t','s','_','6','4','_','2'>, SSATS<DnaMeta<uint64_t, 2> >, boost::mpl::true_, boost::mpl::false_>,
    IndexInfo<boost::mpl::string<'s','s','a','t','s','_','6','4','_','4'>, SSATS<DnaMeta<uint64_t, 4> >, boost::mpl::true_, boost::mpl::false_>
  > set;

  typedef boost::mpl::filter_view<set, boost::mpl::at<boost::mpl::_, boost::mpl::int_<2> > >::type host_set;
  typedef boost::mpl::filter_view<set, boost::mpl::at<boost::mpl::_, boost::mpl::int_<3> > >::type cuda_set;

  template <typename II>
  static IndexSig get_sig() {
    return IndexSig(II::name_c_str::value);
  }
};

// Calls F::apply<I>() where I is the index corresponding to the sig s
template <typename F>
struct IndexDispatch {
  IndexSig sig;
  F & f;
  bool succ;

  IndexDispatch(IndexSig s, F & f) : sig(s), f(f), succ(false) { }

  template <typename II>
  void apply() {
    if (sig == IndexMeta::get_sig<II>()) {
      f.apply<II>();
      succ = true;
    }
  }

  template <typename Set>
  bool dispatch_go() {
    succ = false;
    static_for_each<Set>(*this);
    if (! succ)
      f.failure();
    return succ;
  }

  bool dispatch() {
    return dispatch_go<IndexMeta::set>();
  }

  bool dispatch_host() {
    return dispatch_go<IndexMeta::host_set>();
  }

  bool dispatch_cuda() {
    return dispatch_go<IndexMeta::cuda_set>();
  }
};

} // bmap

#endif // _INDEX_HH
