#include "util/timer_cuda.hh"
#include "index.hh"
#include "ssa_cuda.hh"
#include "index_api.hh"

using namespace bmap;

static const bool print_timings = false;

template <typename KernelIndex, typename kernel_reads_array, typename kernel_ids_array, typename kernel_results_array>
__global__ void find_kernel(
    const KernelIndex index, kernel_reads_array reads, kernel_ids_array read_ids, kernel_results_array results,
    uint offset, uint len, int err_pos, int err, int size) {
  int id = blockIdx.x * blockDim.x + threadIdx.x;
  const int grid_size = gridDim.x * blockDim.x;

  uint sp, ep, count;
  int rid;

  for (; id < size; id += grid_size) {
    rid = read_ids[id];
    reads(rid, offset + err_pos) = (reads(rid, offset + err_pos) + err) % 4;
    count = index.Find(reads[rid], offset, len, sp, ep);
    reads(rid, offset + err_pos) = (reads(rid, offset + err_pos) - err + 4) % 4;
    results[id].sp = sp;
    results[id].count = count;
  }
}

template <typename KernelIndex, typename kernel_reads_array, typename kernel_size_array>
__global__ void compare_to_text_kernel(const KernelIndex index,
    const kernel_reads_array reads, const kernel_size_array reads_map, const kernel_size_array text_pos, int offset, int length, kernel_size_array errors,
    int size) {
  int id = blockIdx.x * blockDim.x + threadIdx.x;
  const int grid_size = gridDim.x * blockDim.x;

  uint pos;

  for (; id < size; id += grid_size) {
    pos = text_pos[id];
    if (pos < offset)
      errors[id] = 999;
    else
      errors[id] = index.CompareToText(pos - offset, length, reads[reads_map[id]]);
  }
}

template <typename KernelIndex, typename kernel_size_array>
__global__ void index_to_text_kernel(const KernelIndex index, const kernel_size_array index_pos, kernel_size_array text_pos, int size) {
  int id = blockIdx.x * blockDim.x + threadIdx.x;
  const int grid_size = gridDim.x * blockDim.x;

  for (; id < size; id += grid_size) {
    text_pos[id] = index.BwtToText(index_pos[id]);
  }
}

template <typename Index>
void CudaIndexAPI<Index>::find(const cuda_strings_type & creads, const cuda_size_array & cread_ids, size_type offset, size_type len, int err_pos, int err, cuda_result_array & cresults) const {
  const int threads = 256;
  const int blocks = 4096;
  CudaTimer tc;

  find_kernel<<<blocks, threads>>>(cuda_index.ForKernel(),
      creads.ForKernel(), cread_ids.ForKernel(), cresults.ForKernel(), offset, len, err_pos, err, cread_ids.size);

  if (print_timings)
    std::cerr << "CUDA find " << cread_ids.size/(1<<10) << "K reads"
      << " took " << tc.time() << "ms [" << cudaGetErrorString(cudaGetLastError()) << "]\n";
}

template <typename Index>
void CudaIndexAPI<Index>::index_to_text(const size_array & index_pos, size_array & text_pos) const {
  CudaTimer tc;

  cuda_size_array cindex_pos = index_pos;
  cuda_size_array ctext_pos(cindex_pos.size);

  index_to_text(cindex_pos, ctext_pos);

  text_pos = ctext_pos;
}

template <typename Index>
void CudaIndexAPI<Index>::index_to_text(const cuda_size_array & index_pos, cuda_size_array & text_pos) const {
  const int threads = 256;
  const int blocks = 4096;

  CudaTimer tc;

  index_to_text_kernel<<<blocks, threads>>>(cuda_index.ForKernel(), index_pos.ForKernel(), text_pos.ForKernel(), index_pos.size);

  if (print_timings)
    std::cerr << "CUDA index_to_text " << index_pos.size/(1<<10) << "K pos took "
      << tc.time() << "ms [" << cudaGetErrorString(cudaGetLastError()) << "]\n";
}

template <typename Index>
void CudaIndexAPI<Index>::compare_to_text(const cuda_strings_type & creads, const cuda_size_array & cread_map,
    const cuda_size_array & ctext_pos, int offset, int length,
    cuda_size_array & cerrors) const {
  const int threads = 256;
  const int blocks = 4096;

  CudaTimer tc;

  compare_to_text_kernel<<<blocks, threads>>>(cuda_index.ForKernel(), creads.ForKernel(), cread_map.ForKernel(), ctext_pos.ForKernel(), offset, length, cerrors.ForKernel(), ctext_pos.size);

  if (print_timings)
    std::cerr << "CUDA compare_to_text " << ctext_pos.size/(1<<10) << "K pos"
      << " took " << tc.time() << "ms [" << cudaGetErrorString(cudaGetLastError()) << "]\n";
}

template class CudaIndexAPI<SSAT<DnaMeta<uint32_t, 3, 8> > >;
template class CudaIndexAPI<SSAT<DnaMeta<uint32_t, 3, 16> > >;
template class CudaIndexAPI<SSAT<DnaMeta<uint32_t, 3, 32> > >;
template class CudaIndexAPI<SSAT<DnaMeta<uint32_t, 3, 64> > >;
