#include <fstream>
#include <string>

#include "util/serialization.hh"
#include "util/timer.hh"
#include "bwt.hh"
#include "ssa.hh"
#include "index.hh"

using namespace bmap;

struct Indexer {
  dna::FastaReader fasta_reader;
  bool rebwt;
  std::iostream & bwt_stream;
  std::ostream & index_stream;
  DnaString dna_string;
  bool dna_loaded;

  Indexer(std::istream & fasta_s, bool reb, std::iostream & bwt_s, std::ostream & index_s) :
    fasta_reader{fasta_s},
    rebwt(reb),
    bwt_stream(bwt_s),
    index_stream(index_s),
    dna_loaded{false}
  {}

  void read_dna() {
    if (dna_loaded)
      return;

    fasta_reader.read_dna(dna_string);
    dna::RandomizeN(dna_string);
    dna_loaded = true;
  }

  DnaCompact text() {
    read_dna();
    DnaCompact dc(dna_string);
    dna_string.Clear();
    return dc;
  }

  template <typename IndexInfo>
  void apply() {
    typedef typename IndexInfo::index_type Index;
    typedef typename Index::bwt_type Bwt;
    typedef typename Index::meta_type Meta;

    SArray<dna::SequenceMeta> meta;
    fasta_reader.get_meta(meta);

    Bwt bwt;

    if (rebwt) {
      read_dna();
      bwt = BuildBWT<Meta>(dna_string);
      Save(bwt, bwt_stream);
      std::cerr << "[bwt] built and saved\n";
    } else {
      Load(bwt, bwt_stream);
      std::cerr << "[bwt] loaded\n";
    }

    Index index = SSABuilder<Index>::Build(*this, bwt);
    index.meta = meta;
    IndexSig sig = IndexMeta::get_sig<IndexInfo>();
    Save(sig, index_stream);
    Save(index, index_stream);
  }

  void failure() {

  }
};

int main(int argc, char * argv[]) {
  srand(31337);
  Timer timer;

  if (argc != 6) {
    std::cerr << "Usage: " << argv[0] << " type rebwt fasta-file bwt-file index-file\n";
    return 1;
  }

  std::string type(argv[1]);
  bool rebwt = atoi(argv[2]) == 1;

  std::string fasta_f = argv[3];
  std::ifstream fasta_s{fasta_f};

  if (! fasta_s.good()) {
    std::cerr << "Bad file " << fasta_f << "\n";
    return EXIT_FAILURE;
  }

  std::string bwt_f = argv[4];
  std::fstream bwt_s{bwt_f};

  if (! bwt_s.good()) {
    bwt_s.close();
    bwt_s.open(bwt_f, std::ios::out);
    if (! bwt_s.good()) {
      std::cerr << "Bad file " << bwt_f << "\n";
      return EXIT_FAILURE;
    }
    rebwt = true;
  }

  std::string index_f = argv[5];
  std::ofstream index_s(index_f);

  if (! index_s.good()) {
    std::cerr << "Bad file " << index_f << "\n";
    return EXIT_FAILURE;
  }

  Indexer indexer{fasta_s, rebwt, bwt_s, index_s};
  IndexSig sig{type};
  IndexDispatch<Indexer> dispatcher{sig, indexer};

  if (! dispatcher.dispatch()) {
    std::cerr << "Unknown index type '" << type << "'\n";
    return EXIT_FAILURE;
  }
}

