#include <vector>
#include <fstream>

#include <cstdlib>
#include <cstdio>
#include <ctime>
#include <cstring>

#include <boost/program_options.hpp>

#include "api/SamHeader.h"
#include "api/BamAlignment.h"
#include "api/BamWriter.h"

#include "util/serialization.hh"
#include "util/timer.hh"
#include "dna.hh"
#include "ssa.hh"
#include "mapper.hh"
#include "index.hh"

using std::string;
using std::cout;
using std::cerr;
using std::cin;
using std::vector;

using namespace BamTools;

using namespace bmap;

namespace po = boost::program_options;

struct MapStart {
  int read_size;
  int max_errors;
  int slices;
  int limit;
  int cutoff;
  std::ifstream & index_s;
  dna::FastqReader & fastq_reader;
  std::string & bam_f;

  MapStart(std::ifstream & is, dna::FastqReader & r, std::string & bf) : index_s(is), fastq_reader(r), bam_f(bf) {}

  template <typename IndexInfo>
  void apply() {
    typedef typename IndexInfo::index_type Index;
    Timer t;
    Index index;
    Load(index, index_s);

    BamWriter bam_writer;
    bam_writer.SetCompressionMode(BamWriter::Uncompressed);
    SamHeader sam_header;
    RefVector ref_vector;

    for(size_t i = 0; i < index.meta.size; i++)
    {
      sam_header.Sequences.Add(index.meta[i].label, index.meta[i].size);
      ref_vector.push_back(RefData(index.meta[i].label, index.meta[i].size));
    }

    bam_writer.Open(bam_f, sam_header, ref_vector);

    std::cout << "Loading index took " << t.msecs() << " ms\n";

    Mapper<Index> mapper(index, read_size, max_errors, slices, cutoff);
    mapper.map(fastq_reader, bam_writer, limit);
  }

  void failure() {
  }
};

int main(int argc, char * argv[]) {
#if 0
  if (argc != 8 && argc != 9) {
    std::cerr << "Usage: " << argv[0] << " index-file fastq-file bam-file size max-errors slices cutoff [limit]\n";
    return 1;
  }
#endif
  std::string index_f;
  std::string fastq_f;
  std::string bam_f;

  uint read_size;
  uint errors;
  int limit;
  uint cutoff;
  uint slices;

  po::options_description main_opt("Main options");
  main_opt.add_options()
    ("help", "produce help message")
    ("index,i", po::value<std::string>(&index_f)->required(), "path to the index")
    ("read-size,r", po::value<uint>(&read_size)->required(), "size of the reads")
    ("errors,e", po::value<uint>(&errors)->default_value(0), "maximum number of errors")
    ("limit,l", po::value<int>(&limit)->default_value(-1), "process only that many reads")
    ;

  po::options_description advanced_opt("Advanced options");
  advanced_opt.add_options()
    ("cutoff,c", po::value<uint>(&cutoff)->default_value(100), "cutoff")
    ("slices,s", po::value<uint>(&slices)->default_value(0), "number of slices")
    ;

  po::options_description hidden_opt("Hidden options");
  hidden_opt.add_options()
    ("fastq", po::value<std::string>(&fastq_f)->required(), "fastq")
    ("bam", po::value<std::string>(&bam_f)->required(), "bam")
    ;
  po::positional_options_description p;
  p.add("fastq", 1);
  p.add("bam", 1);

  po::options_description cmdline_opt;
  cmdline_opt.add(main_opt).add(advanced_opt).add(hidden_opt);

  po::options_description display_opt;
  display_opt.add(main_opt).add(advanced_opt);

  po::variables_map vm;
  try {
    po::store(po::command_line_parser(argc, argv).
        options(cmdline_opt).positional(p).run(), vm);
  } catch (po::unknown_option & e) {
    cerr << "Error: " << e.what() << "\n";
    cout << "Usage: " << argv[0] << " [options] reads_fastq output_bam\n";
    cout << display_opt << "\n";
    return 1;
  } catch (po::invalid_option_value & e) {
    cerr << "Error: " << e.what() << "\n";
    cout << "Usage: " << argv[0] << " [options] reads_fastq output_bam\n";
    cout << display_opt << "\n";
    return 1;
  }

  if (vm.count("help")) {
    cout << "Usage: " << argv[0] << " [options] reads_fastq output_bam\n";
    cout << display_opt << "\n";
    return 0;
  }

  try {
    po::notify(vm);
  } catch (po::required_option & e) {
    cerr << "Error: " << e.what() << "\n";
    cout << "Usage: " << argv[0] << " [options] reads_fastq output_bam\n";
    cout << display_opt << "\n";
    return 1;
  }

  std::ifstream reads{fastq_f};
  dna::FastqReader fastqr(reads);

  std::ifstream index_s{index_f};

  MapStart ms(index_s, fastqr, bam_f);

  ms.read_size = read_size;
  ms.max_errors = errors;
  ms.slices = slices;
  ms.cutoff = cutoff;
  ms.limit = limit;

  IndexSig sig;
  Load(sig, index_s);
  IndexDispatch<MapStart> id(sig, ms);
  id.dispatch();
}
