#ifndef _MAPPER_HH
#define _MAPPER_HH

#include <vector>
#include <algorithm>

#include <api/SamHeader.h>
#include <api/BamAlignment.h>
#include <api/BamWriter.h>

#include "config.hh"
#include "util/array.hh"
#include "util/timer.hh"
#include "ssa.hh"
#include "dna.hh"

using std::vector;
using std::cerr;

const int UNMAPPED = 999;

template <typename Index>
struct MappingSlice {
  typename Index::string_type data;
  typename Index::size_type offset;
  typename Index::State state;
  bool rc;

  MappingSlice() : offset(0), rc(false) {}

  void trunc(typename Index::size_type size) {
    if (size >= data.size)
      return;
    data = typename Index::string_type(size, data.data + data.size - size);
    if (size < state.mapped)
      state.mapped = size;
  };
};

template <typename Index>
struct MappingSeed {
  MappingSeed() : error(0) {}

  typename Index::Result result;
  typename Index::size_type size;
  typename Index::size_type offset;
  int error;
  typename Index::size_type error_pos;
};

template <typename Index>
void PrintMapping(const Index & index, typename Index::string_type read[2], const Mapping<Index> & m) {
  DnaString ms = index.Extract(m.from, read[m.rc].size);
  cerr << "  " << dna::Unmap(ms) << "\n  ";
  cerr << dna::Unmap(read[m.rc]) << "\n  ";
  for (size_t i = 0 ; i < read[m.rc].size ; ++i)
    cerr << (read[m.rc][i] != ms[i] ? "^" : " ");
  cerr << "\n";
}

template <class Mapper, size_t max_slices>
struct MapperState {
  MappingSlice<typename Mapper::index_type> slices[2][max_slices];
};

template <typename Index>
class Mapper {
 public:
  typedef Index index_type;
  typedef Mapping<Index> MappingI;
  typedef MappingSlice<Index> MappingSliceI;
  typedef typename Index::string_type string_type;
  typedef typename Index::string_type mstring;
  typedef typename Index::size_type size_type;
  typedef MapperState<Mapper, 16> State;

  Mapper(Index & index, int rs, int me, int slices, int cutoff);

  void map(dna::FastqReader & reader, BamTools::BamWriter & bam, int limit);

  MappingI backmap_both(mstring read[2]);
  bool backmap_slice(MappingSliceI & slice, MappingI & mapping, mstring & read, bool error=false);
  bool backmap_slice_error(MappingSliceI & slice, MappingI & mapping, mstring & read);

  bool verify_slice(const MappingSliceI & slice, MappingI & mapping, mstring & read, bool error);

  bool UpdateMapping(MappingI & mapping, uint errors, size_type from, bool rc, uint count=1);
  bool enough(MappingI & mapping, bool errors) {
    if (errors) {
      if (mapping.errors <= slices)
        return true;
    } else {
      if (mapping.errors == 1)
        return true;
    }
    return false;
  }

  Index & index;
  uint read_size;
  uint slices;
  uint max_errors;
  uint slice_errors;
  uint cutoff;
  uint total_hits;
  uint checked;
 private:
};

template <typename Index>
Mapper<Index>::Mapper(Index & i, int rs, int me, int s, int c) :
  index(i),
  read_size(rs), slices(s),
  max_errors(me), slice_errors(me / s),
  cutoff(c),
  total_hits(0), checked(0)
{

  cerr << "Mapper: rs=" << read_size << " me=" << max_errors
    << " se=" << slice_errors << " s=" << slices << "\n";
}

template <typename Index>
void errorify(Read<Index> & read, int max_e) {
  read.errors = rand() % (max_e + 1);

  for (int i = 0 ; i < read.errors ; ++i) {
    int pos = rand() % read.data.size;
    read.data[pos] = (read.data[pos] + 1 + pos % 3) % 4;
  }
}

struct Stats {
  int count;
  int skipped;
  int mapped[2];
  int mapped_uniq[2];
  int hits[2];
  int errors[32];
  int max_errors;

  Stats(int me) :
    count(0), skipped(0),
    mapped{0, 0},
    mapped_uniq{0, 0},
    hits{0, 0},
    max_errors(me)
  {
    memset(errors, 0, sizeof(errors));
  }

  int mapped_total() const { return mapped[0] + mapped[1]; }
  int mapped_uniq_total() const { return mapped_uniq[0] + mapped_uniq[1]; }
  int hits_total() const { return hits[0] + hits[1]; }
  double mapped_part() const { return double(mapped_total()) / count; }
  double mapped_uniq_part() const { return double(mapped_uniq_total()) / count; }
  double skipped_part() const { return double(skipped) / count; }
};

void print_stats(const Stats & stats, int time_ms)
{
  cerr << "Mapped (" <<  stats.mapped[0] << " + " << stats.mapped[1]
    << " = " << stats.mapped_total() << ") / " << stats.count
    << " (" << stats.mapped_part() << ")"
    << " in " << time_ms << " ms\n"
    << "  uniq (" <<  stats.mapped_uniq[0] << " + " << stats.mapped_uniq[1]
    << " = " << stats.mapped_uniq_total() << ") / " << stats.count
    << " (" << stats.mapped_uniq_part() << ")\n"
    << "  hits " << stats.hits[0] << " + " << stats.hits[1] << " = " << stats.hits_total() << "\n"
    << "  skip " << stats.skipped << " (" << stats.skipped_part() << ")\n";
  cerr << " errors:";
  for (int i = 0 ; i <= stats.max_errors ; ++i)
    cerr << " " << stats.errors[i];
  cerr << "\n";
}

void print_mini_stats(const Stats & stats, int time_ms)
{
  cerr << "\r" << std::string(120, ' ') << "\r";
  cerr << "Mapped " << stats.mapped_total() << " / " << stats.count
    << " (" << stats.mapped_part() << ")" << " in " << time_ms << " ms" << std::flush;
}

template <typename Index>
void Mapper<Index>::map(dna::FastqReader & fastqr, BamTools::BamWriter & bam_writer, int limit)
{
  Timer t;
  dna::DnaRead read;
  MappingI m;
  Stats stats{max_errors};

  while (++stats.count != limit + 1) {
    if (!fastqr.Next(read))
      break;

    BamTools::BamAlignment ba;

    ba.Name = read.name;
    ba.Length = read.data[0].size;
    ba.QueryBases = dna::Unmap(read.data[0]);
    ba.Qualities = std::string(ba.Length, '2');
    ba.MapQuality = 0;

    if (dna::HasN(read.data[0])) {
      ba.SetIsMapped(false);
      bam_writer.SaveAlignment(ba);
      ++stats.skipped;
      continue;
    }

    // reverse complement
    read.data[1] = read.data[0];
    dna::ReverseComplement(read.data[1]);

    m = backmap_both(read.data);
    if (m.errors <= max_errors) {
      ba.AddTag("NM", "i", m.errors);
      ++stats.mapped[m.rc];
      if (m.count == 1)
        ++stats.mapped_uniq[m.rc];
      ++stats.errors[m.errors];
      stats.hits[m.rc] += m.count;
      //if (stats.mapped_total() % 1000 == 0) {
      //  PrintMapping(ssa, read, m);
      //}
      ba.SetIsMapped(true);
      size_t ref_from = 0;
      size_t ref_id = 0;
      index.get_relative_pos(m.from, ref_from, ref_id);
      ba.Position = ref_from;
      ba.RefID = ref_id;
      ba.SetIsReverseStrand(m.rc);
    } else {
      ba.SetIsMapped(false);
    }
    bam_writer.SaveAlignment(ba);

    if (stats.count % 10000 == 0) {
      print_mini_stats(stats, t.total_msecs());
    }
  }
  --stats.count;
  cerr << "\r" << std::string(120, ' ') << "\r";
  print_stats(stats, t.total_msecs());
  std::cerr << stats.mapped_total() << " " << stats.mapped_uniq_total() << " " << t.total_msecs() / 1000 << " " << stats.count << "\n";
  std::cout << read_size << " " << max_errors << " " << slices
    << " completed " << stats.count << " in " << t.total_msecs() << "ms" << std::endl;

}

template <typename Index>
inline bool Mapper<Index>::UpdateMapping(MappingI & mapping, uint errors, size_type from, bool rc, uint count)
{
  if (errors < mapping.errors) {
    //if (mapping.errors <= max_errors)
    //  cerr << mapping.errors << " -> " << errors << "\n";
    mapping.rc = rc;
    mapping.errors = errors;
    mapping.from = from;
    mapping.count = count;
    return true;
  } else if (errors == mapping.errors && from != mapping.from) {
    mapping.count += count;
    return true;
  }
  return false;
}

template <typename Index>
typename Mapper<Index>::MappingI Mapper<Index>::backmap_both(mstring read[2]) {
  MappingI mapping;
  State state;

  mapping.errors = max_errors + 1;

  bool updated;
  state.slices[0][0].data = read[0];
  updated = backmap_slice(state.slices[0][0], mapping, read[0]);
  if (updated)
    return mapping;
  state.slices[1][0].data = read[1];
  state.slices[1][0].rc = true;
  updated = backmap_slice(state.slices[1][0], mapping, read[1]);
  if (updated)
    return mapping;

  if (max_errors == 0)
    return mapping;

  size_type to_map = read[0].size;
  size_type slices_left = std::min(slices, max_errors + 1);
  auto size = ceildiv(to_map, slices_left--);
  auto offset = to_map - size;

  state.slices[0][0].trunc(size);
  state.slices[0][0].offset = offset;
  state.slices[1][0].trunc(size);
  state.slices[1][0].offset = offset;
  to_map = offset;

  auto i = 1u;
  while (to_map) {
    size = ceildiv(to_map, slices_left--);
    offset = to_map - size;
    for (auto j = 0u ; j < 2 ; ++j) {
      state.slices[j][i].data = mstring(size, read[j].data + offset);
      state.slices[j][i].offset = offset;
      state.slices[j][i].rc = j == 1;
    }

    to_map = offset;
    ++i;
  }

  updated = false;
  for (auto i = 0u ; i < slices ; ++i) {
    updated |= backmap_slice(state.slices[0][i], mapping, read[0]);
    if (enough(mapping, false))
      return mapping;
    updated |= backmap_slice(state.slices[1][i], mapping, read[1]);
    if (enough(mapping, false))
      return mapping;
  }

  bool errors = ((updated ? mapping.errors : max_errors) / slices) > 0;
  if (!errors)
    return mapping;

  for (auto i = 0u; i < slices; ++i) {
    backmap_slice_error(state.slices[0][i], mapping, read[0]);
    if (enough(mapping, true))
      return mapping;
    backmap_slice_error(state.slices[1][i], mapping, read[1]);
    if (enough(mapping, true))
      return mapping;
  }

  return mapping;
}

template <typename Index>
bool Mapper<Index>::backmap_slice(MappingSliceI & slice, MappingI & mapping, mstring & read, bool error)
{
  bool full = index.Find(slice.data, slice.state);

  if (! full)
    return false;

  if (slice.data.size == read.size) {
    UpdateMapping(mapping, error ? 1 : 0, index.Locate(slice.state, 1)[0], slice.rc, slice.state.count);
    return true;
  }

  if (slice.state.count > cutoff)
    return false;

  return verify_slice(slice, mapping, read, error);
}

template <typename Index>
bool Mapper<Index>::backmap_slice_error(MappingSliceI & slice, MappingI & mapping, mstring & read)
{
  size_type i = std::max(slice.data.size - slice.state.mapped - 1, 0U);
  for ( ; i < slice.data.size ; ++i) {
    for (int s = 0 ; s < 3 ; ++s) {
      slice.data[i] = (slice.data[i] + 1) % 4;
      slice.state.mapped = std::min(slice.state.mapped, slice.data.size - i - 1);
      backmap_slice(slice, mapping, read, true);
    }
    slice.data[i] = (slice.data[i] + 1) % 4;
  }
  return false;
}

template <typename Index>
bool Mapper<Index>::verify_slice(const MappingSliceI & slice, MappingI & mapping, mstring & read, bool error)
{
  size_type missing_left = slice.offset;

//TODO alignment
#if 0
  if (missing_left != 0)
    missing_left += max_errors;
#endif

  size_type missing_right = read.size - (slice.data.size + slice.offset);

//TODO alignment
#if 0
  if (missing_right != 0)
    missing_right += max_errors;
#endif

  mstring s(slice.data.size + missing_left + missing_right);

  std::copy(slice.data.data, slice.data.data + slice.data.size, s.data + missing_left);

  //std::string ss = dna::Unmap(read);
  //cerr << "ml " << missing_left << " mr " << missing_right << " seed " << slice.size << " ss " << s.size << "\n";
  //cerr << "Seed:        " << std::string(missing_left, ' ') << ss.substr(slice.offset, slice.data.size) << "\n";

  bool updated = false;
  typename Index::Result result = slice.state.result();

  for (auto i = 0u ; i < result.count ; ++i) {
    auto text_from = index.ExtractLocate(result.sp + i, missing_left, s.data);
    if (text_from + slice.data.size + missing_right > index.size)
      continue;
    if (missing_right)
      index.Extract(text_from + slice.data.size, missing_right, s.data + missing_left + slice.data.size);
    //cerr << " expanded to " << dna::Unmap(s) << " " << text_from << "\n";
    uint e = 0;
    for (auto i = 0u ; e <= mapping.errors && i < read.size ; ++i)
      if (s[i] != read[i])
        ++e;
    if (e <= max_errors)
      updated |= UpdateMapping(mapping, e, text_from - missing_left + 1, slice.rc);
    if (enough(mapping, error))
      return updated;
  }
  return updated;
}


#endif // _MAPPER_HH
