#include <vector>
#include <fstream>
#include <limits>

#include <cstdlib>
#include <cstdio>
#include <ctime>
#include <cstring>

#include "api/SamHeader.h"
#include "api/BamAlignment.h"
#include "api/BamWriter.h"

#include "util/serialization.hh"
#include "util/timer.hh"
#include "dna.hh"
#include "ssa.hh"
#include "mapper_cuda.hh"
#include "index.hh"

using std::string;
using std::cout;
using std::cerr;
using std::cin;
using std::vector;

using namespace BamTools;

using namespace bmap;

// Starter used in the static dispatch on index type
struct MapStart {
  int read_size;
  int max_errors;
  int slices;
  size_t limit;
  std::ifstream & index_s;
  dna::FastqReader & fastq_reader;
  std::string & bam_f;

  MapStart(std::ifstream & is, dna::FastqReader & r, std::string & bf) : index_s(is), fastq_reader(r), bam_f(bf) {}

  template <typename IndexInfo>
  void apply() {
    Timer t;
    typedef typename IndexInfo::index_type Index;
    Index index;
    // Load the index itself
    Load(index, index_s);

    BamWriter bam_writer;
    // Compressions slows down a lot
    bam_writer.SetCompressionMode(BamWriter::Uncompressed);
    SamHeader sam_header;
    RefVector ref_vector;

    // Load chromosomes to bamtools
    for (size_t i = 0; i < index.meta.size; i++) {
      sam_header.Sequences.Add(index.meta[i].label, index.meta[i].size);
      ref_vector.push_back(RefData(index.meta[i].label, index.meta[i].size));
    }

    bam_writer.Open(bam_f, sam_header, ref_vector);

    std::cout << "Loading index took " << t.msecs() << " ms\n";

    Mapper<Index> mapper(index, read_size, max_errors, slices);

    mapper.map(fastq_reader, bam_writer, limit);
  }

  void failure() {
  }
};

int main(int argc, char * argv[]) {
  if (argc != 7 && argc != 8) {
    std::cerr << "Usage: " << argv[0] << " index-file fastq-file bam-file size max-errors slices [limit]\n";
    return 1;
  }

  std::string index_f(argv[1]);
  std::string fastq_f(argv[2]);
  std::string bam_f(argv[3]);

  std::ifstream reads(fastq_f.c_str());
  dna::FastqReader fastqr(reads);

  std::ifstream index_s(index_f.c_str());

  MapStart ms(index_s, fastqr, bam_f);

  ms.read_size = atoi(argv[4]);
  ms.max_errors = atoi(argv[5]);
  ms.slices = atoi(argv[6]);
  ms.limit = std::numeric_limits<size_t>::max();
  if (argc == 8)
    ms.limit = atoi(argv[7]);

  // Doesn't seem to be making much of a difference, we don't use shared mem at all though
  cudaDeviceSetCacheConfig(cudaFuncCachePreferL1);

  IndexSig sig;
  Load(sig, index_s);
  // dispatch on the signature
  IndexDispatch<MapStart> id(sig, ms);
  if (! id.dispatch_cuda()) {
    std::cerr << "Index not supported on CUDA\n";
    return EXIT_FAILURE;
  }
}
