#ifndef _MAPPER_CUDA_HH
#define _MAPPER_CUDA_HH

#include <vector>
#include <algorithm>

#include <api/SamHeader.h>
#include <api/BamAlignment.h>
#include <api/BamWriter.h>

#include "config.hh"
#include "util/array.hh"
#include "util/timer.hh"
#include "util/timer_cuda.hh"
#include "ssa.hh"
#include "dna.hh"
#include "index_api.hh"

#include <omp.h>

using std::vector;
using std::cerr;

// how many reads should be in a single bundle
const uint BUNDLE_SIZE = 500000;

// if we find more hits than that we ignore them
const uint SPEP_LIMIT = 100;

static const bool print_timings = false;

// Holds all the reads (also reverse complemented), their names and mappings
template <typename Index>
struct Reads {
  SArray<std::string> names;
  DnaStrings data;
  SArray<Mapping<Index> > mappings;
  size_t size_;

  Reads(int size, int read_size) :
    names(size),
    data(2 * size, read_size),
    mappings(size)
  {
  }

  size_t size() const { return size_; }

  void resize(size_t n) { size_ = n; }

  Read<Index> operator[](int n) {
    return {names[n], data[2*n], mappings[n]};
  }
};

template <typename Index>
void PrintMapping(const Index & index, typename Index::string_type read[2], const Mapping<Index> & m) {
  DnaString ms = index.Extract(m.from, read[m.rc].size);
  cerr << "  " << dna::Unmap(ms) << "\n  ";
  cerr << dna::Unmap(read[m.rc]) << "\n  ";
  for (size_t i = 0 ; i < read[m.rc].size ; ++i)
    cerr << (read[m.rc][i] != ms[i] ? "^" : " ");
  cerr << "\n";
}

template <typename Index>
class Mapper {
 public:
  typedef Index index_type;
  typedef Read<Index> read_type;
  typedef Reads<Index> reads_type;
  typedef typename Index::Result result;
  typedef typename Index::string_type string_type;
  typedef typename Index::size_type size_type;
  typedef DnaStrings strings_type;
  typedef CSibling<DnaStrings>::Type cuda_strings_type;
  typedef Mapping<Index> MappingI;
  typedef SArray<result> result_array;
  typedef CudaArray<result> cuda_result_array;
  typedef SArray<size_type> size_array;
  typedef CudaArray<size_type> cuda_size_array;

  Mapper(Index & index, int rs, int me, int max_slices);

  int load_reads(dna::FastqReader & fastqr, reads_type & reads, reads_type & skipped_reads, size_t limit);

  void map(dna::FastqReader & reader, BamTools::BamWriter & bam, size_t limit);

  void map_all(reads_type &, const cuda_strings_type &);
  void map_all_slices(reads_type &, const cuda_strings_type &, const size_array &, size_type offset, size_type len, uint err_pos, uint err);
  void filter_reads(reads_type & reads, size_array & read_ids, bool errors);

  bool UpdateMapping(MappingI & mapping, uint errors, size_type from, bool rc, uint count=1);

  void SaveAlignment(BamTools::BamWriter & bam_writer, const read_type & read);

  bool enough(MappingI & mapping, bool errors) {
    if (errors) {
      //if (mapping.errors <= max_slices)
      if (mapping.errors <= max_errors)
        return true;
    } else {
      if (mapping.errors == 1)
        return true;
    }
    return false;
  }

  Index & index;
  CudaIndexAPI<Index> index_api;
  uint read_size;
  uint max_slices;
  uint max_errors;
  uint slice_errors;
  uint total_hits;
  uint checked;
 private:
};

template <typename Index>
Mapper<Index>::Mapper(Index & i, int rs, int me, int s) :
  index(i),
  index_api(i),
  read_size(rs), max_slices(s),
  max_errors(me), slice_errors(me / s),
  total_hits(0), checked(0)
{

  cerr << "Mapper: rs=" << read_size << " me=" << max_errors
    << " se=" << slice_errors << " s=" << max_slices << "\n";
}

struct Stats {
  int count;
  int skipped;
  int mapped[2];
  int mapped_uniq[2];
  int hits[2];
  int errors[32];
  int max_errors;

  Stats(int me) :
    count(0), skipped(0),
    max_errors(me)
  {
    mapped[0] = mapped[1] = 0;
    mapped_uniq[0] = mapped_uniq[1] = 0;
    hits[0] = hits[1] = 0;
    memset(errors, 0, sizeof(errors));
  }

  int mapped_total() const { return mapped[0] + mapped[1]; }
  int mapped_uniq_total() const { return mapped_uniq[0] + mapped_uniq[1]; }
  int hits_total() const { return hits[0] + hits[1]; }
  double mapped_part() const { return double(mapped_total()) / count; }
  double mapped_uniq_part() const { return double(mapped_uniq_total()) / count; }
  double skipped_part() const { return double(skipped) / count; }
};

void print_stats(const Stats & stats, int time_ms)
{
  cerr << "Mapped (" <<  stats.mapped[0] << " + " << stats.mapped[1]
    << " = " << stats.mapped_total() << ") / " << stats.count
    << " (" << stats.mapped_part() << ")"
    << " in " << time_ms << " ms\n"
    << "  uniq (" <<  stats.mapped_uniq[0] << " + " << stats.mapped_uniq[1]
    << " = " << stats.mapped_uniq_total() << ") / " << stats.count
    << " (" << stats.mapped_uniq_part() << ")\n"
    << "  hits " << stats.hits[0] << " + " << stats.hits[1] << " = " << stats.hits_total() << "\n"
    << "  skip " << stats.skipped << " (" << stats.skipped_part() << ")\n";
  cerr << " errors:";
  for (int i = 0 ; i <= stats.max_errors ; ++i)
    cerr << " " << stats.errors[i];
  cerr << "\n";
}

void print_mini_stats(const Stats & stats, int time_ms)
{
  cerr << "\r" << std::string(120, ' ') << "\r";
  cerr << "Mapped " << stats.mapped_total() << " / " << stats.count
    << " (" << stats.mapped_part() << ")" << " in " << time_ms << " ms" << std::flush;
}

template <typename Index>
void Mapper<Index>::SaveAlignment(BamTools::BamWriter & bam_writer, const read_type & read) {
  BamTools::BamAlignment ba;

  ba.Name = read.name;
  ba.Length = read.data.size();
  ba.QueryBases = dna::Unmap(read.data);
  // FastqReader discards qualities currently
  ba.Qualities = std::string(ba.Length, '2');
  ba.MapQuality = 0;

  bool mapped = read.mapping.count > 0;

  ba.SetIsMapped(mapped);
  if (mapped) {
    size_t ref_from = 0;
    size_t ref_id = 0;
    ba.AddTag("NM", "i", read.mapping.errors);
    ba.SetIsReverseStrand(read.mapping.rc);

    // Get pos relative to a chromosome
    index.get_relative_pos(read.mapping.from, ref_from, ref_id);
    ba.Position = ref_from;
    ba.RefID = ref_id;
  }
  bam_writer.SaveAlignment(ba);
}

// loads up to limit or BUNDLE_SIZE reads
template <typename Index>
int Mapper<Index>::load_reads(dna::FastqReader & fastqr, reads_type & reads, reads_type & skipped_reads, uint64_t limit) {
  uint done = 0;
  uint skipped = 0;

  while (done < BUNDLE_SIZE && skipped < BUNDLE_SIZE && done + skipped < limit) {
    // reset the mapping
    reads.mappings[done] = MappingI();

    bool more = fastqr.Next(reads[done]);
    if (! more)
      break;

    if (dna::HasN(reads.data[done * 2])) {
      // skip reads with N in them
      skipped_reads[skipped] = reads[done];
      ++skipped;
    } else {
      // store the reverse complemented read
      dna::ReverseComplement(reads.data[done * 2], reads.data[done * 2 + 1]);
      ++done;
    }
  }

  reads.resize(done);
  skipped_reads.resize(skipped);

  return done + skipped;
}

template <typename Index>
void Mapper<Index>::map(dna::FastqReader & fastqr, BamTools::BamWriter & bam_writer, size_t limit)
{
  Timer t;
  RealTimer real_time;
  Stats stats(max_errors);

  int reading_ms = 0;
  int mapping_ms = 0;
  int saving_ms = 0;

// Use threads to make sure the GPU (mapping phase) is busy as much as possible
// When there is a lot of work for the GPU it won't use much more than a single core
#pragma omp parallel shared(stats, fastqr, bam_writer, limit) num_threads(3)
  {
    reads_type reads(BUNDLE_SIZE, read_size);
    reads_type skipped_reads(BUNDLE_SIZE, read_size);
    int tid = omp_get_thread_num();

    bool end;
    while (true) {
#pragma omp critical (reading)
      {
        CpuTimer rt;
        end = load_reads(fastqr, reads, skipped_reads, limit) == 0;
        limit -= reads.size() + skipped_reads.size();
        if (!end && print_timings)
          std::cerr << tid << " reading took " << rt.msecs() << "ms\n";
        reading_ms += rt.total_msecs();
      }
      if (end)
        break;

// TODO not really critical but helps; remove when introducing cuda streams
#pragma omp critical (mapping)
      {
        CpuTimer rt;
        cuda_strings_type cdna_strings = reads.data;
        if (reads.size() != cdna_strings.size())
          cdna_strings.Resize(reads.size() * 2);
        map_all(reads, cdna_strings);
        if (print_timings)
          std::cerr << tid << " mapping took " << rt.msecs() << "ms\n";
        mapping_ms += rt.total_msecs();
      }

#pragma omp critical (saving)
      {
        CpuTimer rt;
        for (size_type i = 0; i < skipped_reads.size(); i++)
          SaveAlignment(bam_writer, skipped_reads[i]);

        for (size_type i = 0; i < reads.size(); i++) {
          SaveAlignment(bam_writer, reads[i]);

          // update stats
          MappingI & m = reads.mappings[i];
          if (m.count > 0) {
            ++stats.mapped[m.rc];
            if (m.count == 1)
              ++stats.mapped_uniq[m.rc];
            ++stats.errors[m.errors];
            stats.hits[m.rc] += m.count;
          }
        }
        stats.count += reads.size() + skipped_reads.size();
        stats.skipped += skipped_reads.size();
        if (print_timings)
          std::cerr << tid << " saving took " << rt.msecs() << "ms\n";
        saving_ms += rt.total_msecs();
      }
      if (! print_timings)
        print_mini_stats(stats, real_time.total_msecs());
    }
  }
  // Clear the line
  if (! print_timings)
    cerr << "\r" << std::string(120, ' ') << "\r";

  print_stats(stats, real_time.total_msecs());

  // times
  cerr << "CPU times reading " << reading_ms << " ms ";
  cerr << "mapping " << mapping_ms << " ms ";
  cerr << "saving " << saving_ms << " ms\n";
  cerr << stats.mapped_total() << " " << stats.mapped_uniq_total() << " " << real_time.total_msecs() / 1000 << " " << stats.count << "\n";
}

// update mapping with a new hit
template <typename Index>
inline bool Mapper<Index>::UpdateMapping(MappingI & mapping, uint errors, size_type from, bool rc, uint count)
{
  if (errors > max_errors)
    return false;
  if (errors < mapping.errors) {
    // A new better mapping
    mapping.rc = rc;
    mapping.errors = errors;
    mapping.from = from;
    mapping.count = count;
  } else if (errors == mapping.errors && from != mapping.from) {
    // another hit with the same number of errors
    // only a single mapping is stored currently, but it is remembered that it's not unique
    mapping.count += count;
  }
  return true;
}

// filter out already mapped reads
template <typename Index>
void Mapper<Index>::filter_reads(reads_type & reads, size_array & read_ids, bool errors)
{
  int j = 0;
  for (uint i = 0; i < read_ids.size; i += 2) {
    if (! enough(reads[read_ids[i] / 2].mapping, errors)) {
      read_ids[j++] = read_ids[i];
      read_ids[j++] = read_ids[i] + 1;
    }
  }

  read_ids.size = j;
}

// map all reads
template <typename Index>
void Mapper<Index>::map_all(reads_type & reads, const cuda_strings_type & cdna_strings) {
  size_array read_ids(cdna_strings.size());
  for (uint i = 0; i < read_ids.size; ++i)
    read_ids[i] = i;

  cuda_size_array cread_ids = read_ids;

  // map all reads with no errors
  map_all_slices(reads, cdna_strings, cread_ids, 0, read_size, 0, 0);

  // only exact match?
  if (max_errors == 0)
    return;

  filter_reads(reads, read_ids, false);
  cread_ids = read_ids;

  //cerr << "After exact filtered down to " << cread_ids.size / (1<<10) << "K\n";

  // dividing into slices (fragments of the read)
  size_type to_map = read_size;
  size_type slices_left = std::min(max_slices, max_errors + 1);

  while (to_map) {
    size_type size = ceildiv(to_map, slices_left--);
    size_type offset = to_map - size;

    // map all slices with a specific offset (and size)
    map_all_slices(reads, cdna_strings, cread_ids, offset, size, 0, 0);

    to_map = offset;
  }

  // need error generation?
  if (max_errors / max_slices == 0)
    return;

  filter_reads(reads, read_ids, true);
  cread_ids = read_ids;
  //cerr << "After slices filtered down to " << cread_ids.size / (1<<10) << "K\n";

  to_map = read_size;
  slices_left = max_slices;

  while (to_map) {
    size_type size = ceildiv(to_map, slices_left--);
    size_type offset = to_map - size;

    for (uint err_pos = 0; err_pos < size; ++err_pos) {
      for (uint err = 1; err < 4; ++err) {
        // map all slices with a specific offset (and size) and err at err_pos
        map_all_slices(reads, cdna_strings, cread_ids, offset, size, err_pos, err);
      }
      filter_reads(reads, read_ids, true);
      cread_ids = read_ids;
      //cerr << "filtered down to " << cread_ids.size / (1<<10) << "K\n";
    }

    to_map = offset;
  }
}

template <typename Index>
void Mapper<Index>::map_all_slices(reads_type & reads, const cuda_strings_type & cdna_strings, const size_array & read_ids,
    size_type offset, size_type len, uint err_pos, uint err)
{
  result_array results;
  cuda_size_array cread_ids = read_ids;
  cuda_result_array cresults(read_ids.size);

  // launch the kernel
  index_api.find(cdna_strings, cread_ids, offset, len, err_pos, err, cresults);
  results = cresults;

  size_type spep_count = 0;

  // calculate the size of all results (i.e. the sum of the numbers of hits for each slice)
  for (size_type i = 0; i < results.size; i++) {
    if (results[i].count > SPEP_LIMIT)
      // cut-off too much work
      results[i].count = 1;
      //results[i].count = SPEP_LIMIT;

    spep_count += results[i].count;
  }

  size_array text_from(spep_count);
  size_array read_map(spep_count);
  spep_count = 0;
  for (size_type i = 0; i < results.size; i++) {
    const result & r = results[i];
    for (size_type j = 0; j < r.count; j++) {
      // index position to be translated
      text_from[spep_count] = r.sp + j;
      // id of the corresponding read for later verification
      read_map[spep_count] = read_ids[i];
      ++spep_count;
    }
  }
  cuda_size_array ctext_from = text_from;
  // convert all index positions to text positions
  index_api.index_to_text(ctext_from, ctext_from);
  text_from = ctext_from;

  size_array errors;

  // verify only if matching a slice (not the whole read)
  if (len != read_size) {
    cuda_size_array cerrors(ctext_from.size);
    cuda_size_array cread_map = read_map;
    index_api.compare_to_text(cdna_strings, cread_map, ctext_from, offset, read_size, cerrors);
    errors = cerrors;
  }

  // update the mappings after verification
  for (size_type i = 0; i < read_map.size; ++i) {
    size_type readi = read_map[i];
    bool rc = readi % 2 == 1;

    if (len == read_size)
      UpdateMapping(reads.mappings[readi / 2], 0, text_from[i], rc);
    else
      UpdateMapping(reads.mappings[readi / 2], errors[i], text_from[i] - offset, rc);
  }
}

#endif // _MAPPER_CUDA_HH
