#ifndef _SSA_HH
#define _SSA_HH 1

#include <iostream>

#include <boost/static_assert.hpp>
#include <boost/serialization/serialization.hpp>
#include <boost/serialization/split_member.hpp>
#include <boost/serialization/is_bitwise_serializable.hpp>

#include "config.hh"
#include "util/array.hh"
#include "util/misc.hh"
#include "util/timer.hh"
#include "bits.hh"
#include "bwt.hh"
#include "index_base.hh"
#include "dna.hh"

#include "ssa_base.hh"

template <typename Meta>
struct SpEp {
  typedef typename Meta::size_type size_type;

  SpEp() : sp(0), ep(0) {}
  SpEp(size_type s, size_type e) : sp(s), ep(e) {}

  size_type sp, ep;
};

namespace boost {
namespace serialization {
template <typename T>
struct is_bitwise_serializable<SpEp<T> > : mpl::true_ {};
}}

template <typename Meta, class SSAImp>
class SSABase {
 public:
  typedef SSABaseSiblings<Meta, SSAImp> SiblingTypes;

  typedef Meta meta_type;
  typedef BWT<Meta> bwt_type;
  typedef typename Meta::size_type size_type;
  typedef typename Meta::size_array size_array;
  typedef typename Meta::char_type char_type;
  typedef typename Meta::string_type string_type;
  typedef typename Meta::compact_type compact_type;
  typedef typename Meta::wt_rank_type wt_rank_type;
  static const char_type unknown_char = Meta::unknown_char;
  static const int sigma_size = Meta::sigma_size;
  static const int sample_rate_text = Meta::sample_rate_text;
  static const int sample_rate_bwt = Meta::sample_rate_bwt;
  typedef MatchingState<SSABase, Meta::state_size> State;
  typedef MatchingResult<SSABase> Result;
  typedef SArray<dna::SequenceMeta> meta_array;
  static const int spep_cache_size = Meta::spep_cache_size;
  typedef SArray<SpEp<Meta> > spep_cache_type;

  BOOST_STATIC_ASSERT(sigma_size == 4);

  SSABase() {};

  void Build(const BWT<Meta> & bwt);

  char_type Bwt(size_type n) const;
  char_type BwtOcc(size_type n, size_type & r) const;

  size_type LF(size_type p) const;

  size_type C(char_type c) const;
  size_type Occ(char_type c, size_type k) const;
  size_type OccRank(char_type c, size_type k) const;

  size_type Count(const string_type & pattern) const;
  size_array Locate(const string_type & pattern) const;
  size_array Locate(State & state) const;
  size_array Locate(State & state, size_type count) const;
  size_array Locate(State & state, size_type count, size_type offset) const;

  size_type Find(const string_type & pattern, size_type & sp, size_type & ep) const;
  bool Find(const string_type & pattern, State & state) const;

  void get_relative_pos(size_t pos_in, size_t & pos_out, size_t & id);
  void build_spep_cache();
  void get_cached_spep(const string_type & pattern, size_type & sp, size_type & ep) const;

  size_type size;

  spep_cache_type spep_cache;

 //protected:
  size_type bwt_end_index;

  size_type smaller_than[sigma_size + 1];

  wt_rank_type wt_root;
  wt_rank_type wt_leaves[2];

  meta_array meta;

  friend class boost::serialization::access;
  template <typename Archive>
  void serialize(Archive & ar, const unsigned int version);
};

template <typename Meta>
class SSA : public SSABase<Meta, SSA<Meta> > {
 public:
  typedef SSASiblings<Meta> SiblingTypes;

  typedef typename Meta::size_type size_type;
  typedef typename Meta::size_array size_array;
  typedef typename Meta::char_type char_type;
  typedef typename Meta::string_type string_type;
  typedef typename Meta::compact_type compact_type;
  typedef typename Meta::sampled_bwt_rank_type sampled_bwt_rank_type;
  using SSABase<Meta, SSA<Meta> >::unknown_char;

  using SSABase<Meta, SSA<Meta> >::size;
  using SSABase<Meta, SSA<Meta> >::bwt_end_index;
  using SSABase<Meta, SSA<Meta> >::Locate;

  using SSABase<Meta, SSA<Meta> >::sample_rate_text;
  using SSABase<Meta, SSA<Meta> >::sample_rate_bwt;

  SSA(){};

  void Build(const BWT<Meta> & bwt);

  string_type Extract(size_type from, size_type length) const;
  void Extract(size_type text_from, size_type length, char_type *to) const;

  size_array Locate(size_type sp, size_type ep) const;
  size_type ExtractLocate(size_type sp, size_type length, char_type *to) const;

  size_type BwtToText(size_type n) const;
  size_type TextToBwt(size_type n) const;

 //protected:
  void SamplePos();

  size_array text_to_bwt;
  sampled_bwt_rank_type sampled_bwt;
  size_array bwt_to_text;

  friend class boost::serialization::access;
  template <typename Archive>
  void serialize(Archive & ar, const unsigned int version);
};

template <typename Meta>
class SSAT : public SSABase<Meta, SSAT<Meta> > {
 public:
  typedef SSATSiblings<Meta> SiblingTypes;

  typedef typename Meta::size_type size_type;
  typedef typename Meta::size_array size_array;
  typedef typename Meta::char_type char_type;
  typedef typename Meta::string_type string_type;
  typedef typename Meta::compact_type compact_type;
  typedef typename Meta::sampled_bwt_rank_type sampled_bwt_rank_type;
  using SSABase<Meta, SSAT<Meta> >::unknown_char;

  using SSABase<Meta, SSAT<Meta> >::size;
  using SSABase<Meta, SSAT<Meta> >::bwt_end_index;
  using SSABase<Meta, SSAT<Meta> >::Locate;

  using SSABase<Meta, SSAT<Meta> >::sample_rate_bwt;

  SSAT(){};

#ifdef __CXX0X__
  void Build(const BWT<Meta> & bwt, compact_type && text);
#else
  void Build(const BWT<Meta> & bwt, const compact_type & text);
#endif

  string_type Extract(size_type from, size_type length) const;
  void Extract(size_type text_from, size_type length, char_type *to) const;

  size_array Locate(size_type sp, size_type ep) const;
  size_type ExtractLocate(size_type sp, size_type length, char_type *to) const;

  size_type BwtToText(size_type n) const;
  size_type TextToBwt(size_type n) const;

 //protected:
  void SamplePos();

  compact_type text;

  sampled_bwt_rank_type sampled_bwt;
  size_array bwt_to_text;

  friend class boost::serialization::access;
  template <typename Archive>
  void serialize(Archive & ar, const unsigned int version);
};

template <typename Meta>
class SSATS : public SSABase<Meta, SSATS<Meta> > {
 public:
  typedef typename Meta::size_type size_type;
  typedef typename Meta::size_array size_array;
  typedef typename Meta::char_type char_type;
  typedef typename Meta::string_type string_type;
  typedef typename Meta::compact_type compact_type;
  using SSABase<Meta, SSATS<Meta> >::unknown_char;

  using SSABase<Meta, SSATS<Meta> >::size;
  using SSABase<Meta, SSATS<Meta> >::bwt_end_index;
  using SSABase<Meta, SSATS<Meta> >::Locate;

  SSATS(){};

#ifdef __CXX0X__
  void Build(const BWT<Meta> & bwt, compact_type && text);
#endif

  string_type Extract(size_type from, size_type length) const;
  void Extract(size_type text_from, size_type length, char_type *to) const;

  size_array Locate(size_type sp, size_type ep) const;
  size_type ExtractLocate(size_type sp, size_type length, char_type *to) const;

  size_type BwtToText(size_type n) const;
  size_type TextToBwt(size_type n) const;

 //protected:
  void SamplePos();

  CompactArray<unsigned long, 2, char_type> text;

  size_array bwt_to_text;

  friend class boost::serialization::access;
  template <typename Archive>
  void serialize(Archive & ar, const unsigned int version);
};

template <typename Meta, typename SSAImp>
template <typename Archive>
void SSABase<Meta, SSAImp>::serialize(Archive & ar, const unsigned int version) {
  ar & size;
  ar & bwt_end_index;
  ar & smaller_than;

  ar & wt_root;
  ar & wt_leaves[0];
  ar & wt_leaves[1];

  ar & meta;

  ar & spep_cache;
}

template <typename Meta>
template <typename Archive>
void SSA<Meta>::serialize(Archive & ar, const unsigned int version) {
  SSABase<Meta, SSA<Meta> >::serialize(ar, version);

  ar & sampled_bwt;
  ar & bwt_to_text;
  ar & text_to_bwt;
}

template <typename Meta>
template <typename Archive>
void SSAT<Meta>::serialize(Archive & ar, const unsigned int version) {
  SSABase<Meta, SSAT<Meta> >::serialize(ar, version);

  ar & text;

  ar & sampled_bwt;
  ar & bwt_to_text;
}

template <typename Meta>
template <typename Archive>
void SSATS<Meta>::serialize(Archive & ar, const unsigned int version) {
  SSABase<Meta, SSATS<Meta> >::serialize(ar, version);

  ar & text;

  ar & bwt_to_text;
}

template <typename Meta, class SSAImp>
void SSABase<Meta, SSAImp>::get_relative_pos(size_t pos_in, size_t & pos_out, size_t & id)
{
    pos_out = pos_in;

    for(size_t i = 0; i < meta.size; i++)
    {
      if(pos_out < meta[i].size)
      {
        id = i;
        break;
      }
      else
      {
        pos_out -= meta[i].size;
      }
    }

//    std::cout << pos_in << " -> " << pos_out << " seq:" << seq_name << std::endl;
}


template <typename Meta, class SSAImp>
void SSABase<Meta, SSAImp>::build_spep_cache()
{
  if (spep_cache_size == 0)
    return;

  Timer timer;

  int spep_size = 1 << (spep_cache_size * 2);
  std::cerr << "SPEP Cache size: " << (spep_size * sizeof(SpEp<Meta>)) / (1 << 20) << "MB"<< std::endl;
  spep_cache.Resize(spep_size);

  size_type sp, ep;
  const int bpc = Meta::bits_per_char;

  string_type pattern;
  pattern.Resize(spep_cache_size);

  for (int i = 0; i < spep_size; i++) {
    size_t tmp = i;

    for (int j = 0; j < spep_cache_size; j++) {
      dna_t c = tmp & ~(0xffffffffffff << bpc);
      pattern[j] = c;
      tmp >>= bpc;
    }

    Find(pattern, sp, ep);
    spep_cache[i] = SpEp<Meta>(sp, ep);
  }
  std::cerr << "Building SPEP Cache took " << timer.msecs() << " msecs\n";
}

template <typename Meta, class SSAImp>
void SSABase<Meta, SSAImp>::get_cached_spep(const string_type & pattern, size_type & sp, size_type & ep) const
{
  //get integer representation
  size_t ind = 0;
  for(size_t i = 0; i < spep_cache_size; i++) {
    ind <<= Meta::bits_per_char;
    ind |= pattern[pattern.size - i - 1];
  }
  sp = spep_cache[ind].sp;
  ep = spep_cache[ind].ep;
}

template <typename Meta, class SSAImp>
void SSABase<Meta, SSAImp>::Build(const BWT<Meta> & bwt) {
  Timer timer;

  size = bwt.bwt.size - 1;
  bwt_end_index = bwt.end_index;

  for (char_type c = 0 ; c < sigma_size + 1 ; ++c)
    smaller_than[c] = 0;

  for (size_type i = 0 ; i < size + 1 ; ++i)
    ++smaller_than[bwt.bwt[i]];

  wt_root = wt_rank_type(size + 1);
  wt_leaves[0] = wt_rank_type(smaller_than[0] + smaller_than[1]);
  wt_leaves[1] = wt_rank_type(smaller_than[2] + smaller_than[3]);

  for (char_type c = 0 ; c < sigma_size + 1 ; ++c)
    smaller_than[c] = 0;

  for (size_type i = 0 ; i < size + 1 ; ++i) {
    char_type c = bwt.bwt[i];
    if (c == 1) {
        wt_leaves[0].Set(smaller_than[0] + smaller_than[1]);
    } else if (c >= 2) {
      wt_root.Set(i);
      if (c == 3)
        wt_leaves[1].Set(smaller_than[2] + smaller_than[3]);
    }
    ++smaller_than[c];
  }
  std::cerr << "WT setting bits took " << timer.msecs() << " msecs\n";

  wt_root.CalculateRanks();
  wt_leaves[0].CalculateRanks();
  wt_leaves[1].CalculateRanks();

  std::cerr << "WT calculating ranks took " << timer.msecs() << " msecs\n";

  for (char_type c = 1 ; c < sigma_size + 1 ; ++c) {
    smaller_than[c] += smaller_than[c - 1];
  }
  for (char_type c = sigma_size - 1 ; c > 0 ; --c) {
    smaller_than[c] = smaller_than[c - 1];
  }
  smaller_than[0] = 1;
}

template <typename Meta>
void SSA<Meta>::Build(const BWT<Meta> & bwt)
{
  SSABase<Meta, SSA<Meta> >::Build(bwt);

  sampled_bwt = size + 1;
  bwt_to_text = SArray<size_type>(ceildiv(size + 1, sample_rate_bwt));
  text_to_bwt = SArray<size_type>(ceildiv(size, sample_rate_text) + 1);

  SamplePos();
  SSABase<Meta, SSA<Meta> >::build_spep_cache();
}

#ifdef __CXX0X__
template <typename Meta>
void SSAT<Meta>::Build(const BWT<Meta> & bwt, compact_type && t)
{
  SSABase<Meta, SSAT<Meta> >::Build(bwt);

  text = std::move(t);
  sampled_bwt = size + 1;
  bwt_to_text = SArray<size_type>(ceildiv(size + 1, sample_rate_bwt));

  SamplePos();
  SSABase<Meta, SSAT<Meta> >::build_spep_cache();
}

template <typename Meta>
void SSATS<Meta>::Build(const BWT<Meta> & bwt, compact_type && t)
{
  SSABase<Meta, SSATS<Meta> >::Build(bwt);

  text = std::move(t);
  bwt_to_text = SArray<size_type>(size + 1);

  SamplePos();
  SSABase<Meta, SSATS<Meta> >::build_spep_cache();
}
#else
template <typename Meta>
void SSAT<Meta>::Build(const BWT<Meta> & bwt, const compact_type & t)
{
  SSABase<Meta, SSAT<Meta> >::Build(bwt);

  text = t;
  sampled_bwt = size + 1;
  bwt_to_text = SArray<size_type>(ceildiv(size + 1, sample_rate_bwt));

  SamplePos();
  SSABase<Meta, SSAT<Meta> >::build_spep_cache();
}
#endif

template <typename Meta, class SSAImp>
inline typename Meta::char_type SSABase<Meta, SSAImp>::Bwt(size_type n) const {
  size_type r;
  return BwtOcc(n, r);
}

template <typename Meta, class SSAImp>
inline typename Meta::char_type SSABase<Meta, SSAImp>::BwtOcc(size_type n, size_type & r) const {
  if (! wt_root[n]) {
    r = wt_root.Rank0(n) - 1;
    if (! wt_leaves[0][r]) {
      r = wt_leaves[0].Rank0(r);
      if (n > bwt_end_index)
        --r;
      else if (n == bwt_end_index)
        r = 0;
      return 0;
    } else {
      r = wt_leaves[0].Rank(r);
      return 1;
    }
  } else {
    r = wt_root.Rank(n) - 1;
    if (! wt_leaves[1][r]) {
      r = wt_leaves[1].Rank0(r);
      return 2;
    } else {
      r = wt_leaves[1].Rank(r);
      return 3;
    }
  }
}

template <typename Meta, class SSAImp>
inline typename Meta::size_type SSABase<Meta, SSAImp>::LF(size_type p) const {
  size_type occ;
  char_type c = BwtOcc(p, occ);
  p = C(c) + occ - 1;
  return p;
}

template <typename Meta>
typename Meta::size_type SSA<Meta>::TextToBwt(size_type n) const {
  size_type p = text_to_bwt[ceildiv(n, sample_rate_text)];

  if (p == bwt_end_index) {
    n = size - n;
  } else {
    // Extra mod to handle exact hit
    n = (sample_rate_text - n % sample_rate_text) % sample_rate_text;
  }

  while (n--) {
    p = LF(p);
  }

  return p;
}

template <typename Meta>
typename Meta::size_type SSA<Meta>::BwtToText(size_type n) const {
  size_type dist = 0;

  while (! sampled_bwt[n]) {
    n = LF(n);
    ++dist;
  }

  return bwt_to_text[sampled_bwt.Rank(n) - 1] + dist;
}

template <typename Meta>
typename Meta::size_type SSAT<Meta>::BwtToText(size_type n) const {
  size_type dist = 0;

  while (! sampled_bwt[n]) {
    n = LF(n);
    ++dist;
  }

  return bwt_to_text[sampled_bwt.Rank(n) - 1] + dist;
}

template <typename Meta>
typename Meta::size_type SSATS<Meta>::BwtToText(size_type n) const {
  return bwt_to_text[n];
}

template <typename Meta, class SSAImp>
inline typename Meta::size_type SSABase<Meta, SSAImp>::C(char_type c) const {
  return smaller_than[c];
}

template <typename Meta, class SSAImp>
inline typename Meta::size_type SSABase<Meta, SSAImp>::Occ(char_type c, size_type k) const {
  return OccRank(c, k);
}

template <typename Meta, class SSAImp>
typename Meta::size_type SSABase<Meta, SSAImp>::OccRank(char_type c, size_type k) const {
#define IMP 1
#if IMP == 1
  size_type r;

  if (c < 2) {
    r = wt_root.Rank0(k) - 1;
    if (c == 0) {
      r = wt_leaves[0].Rank0(r);
      if (k >= bwt_end_index)
        --r;
    } else {
      r = wt_leaves[0].Rank(r);
    }
  } else {
    r = wt_root.Rank(k) - 1;
    if (c == 2)
      r = wt_leaves[1].Rank0(r);
    else
      r = wt_leaves[1].Rank(r);
  }
  return r;
#elif IMP == 2
  size_type r = wt_root.RankM(k, c/2) - 1;
  r = wt_leaves[c/2].RankM(r, c & 1);
#if 0
  if (c == 0 && k >= bwt_end_index)
    --r;
#else
  r -= (c == 0) && (k >= bwt_end_index);
#endif

  return r;
#elif IMP == 3
  size_type r;
  switch (c) {
    case 0:
      r = wt_root.Rank0(k) - 1;
      r = wt_leaves[0].Rank0(r);
#if 0
      r -= k >= bwt_end_index;
#else
      if (k >= bwt_end_index)
        --r;
#endif
      break;
    case 1:
      r = wt_root.Rank0(k) - 1;
      r = wt_leaves[0].Rank(r);
      break;
    case 2:
      r = wt_root.Rank(k) - 1;
      r = wt_leaves[1].Rank0(r);
      break;
    case 3:
      r = wt_root.Rank(k) - 1;
      r = wt_leaves[1].Rank(r);
      break;
  }
  return r;
#endif
#undef IMP
}

template <typename Meta>
void SSA<Meta>::SamplePos() {
  size_type count = 0;
  size_type p = 0;

  Timer timer;

  while (++count <= size) {
    if ((size - count) % sample_rate_text == 0) {
      text_to_bwt[(size - count) / sample_rate_text] = p;
    }
    if ((size - count + 1) % sample_rate_bwt == 0) {
      sampled_bwt.Set(p);
    }
    p = LF(p);
  }
  // bwt is +1 in size
  if ((size - count + 1) % sample_rate_bwt == 0) {
    sampled_bwt.Set(p);
  }

  std::cerr << "Sampling phase-1 took " << timer.msecs() << " msecs\n";

  text_to_bwt[text_to_bwt.size - 1] = bwt_end_index;

  sampled_bwt.CalculateRanks();
  count = 0;
  p = 0;

  while (++count <= size + 1) {
    if ((size - count + 1) % sample_rate_bwt == 0) {
      bwt_to_text[sampled_bwt.Rank(p) - 1] = (size - count + 1) % (size + 1);
    }
    p = LF(p);
  }

  std::cerr << "Sampling phase-2 took " << timer.msecs() << " msecs\n";
}

template <typename Meta>
void SSAT<Meta>::SamplePos() {
  size_type count = 0;
  size_type p = 0;

  Timer timer;

  while (++count <= size + 1) {
    if ((size - count + 1) % sample_rate_bwt == 0) {
      sampled_bwt.Set(p);
    }
    p = LF(p);
  }

  std::cerr << "Sampling phase-1 took " << timer.msecs() << " msecs\n";

  sampled_bwt.CalculateRanks();
  count = 0;
  p = 0;

  while (++count <= size + 1) {
    if ((size - count + 1) % sample_rate_bwt == 0) {
      bwt_to_text[sampled_bwt.Rank(p) - 1] = (size - count + 1) % (size + 1);
    }
    p = LF(p);
  }

  std::cerr << "Sampling phase-2 took " << timer.msecs() << " msecs\n";
}

template <typename Meta>
void SSATS<Meta>::SamplePos() {
  size_type count = 0;
  size_type p = 0;

  Timer timer;

  while (++count <= size + 1) {
    bwt_to_text[p] = (size - count + 1) % (size + 1);
    p = LF(p);
  }

  std::cerr << "Sampling took " << timer.msecs() << " msecs\n";
}

template <typename Meta, class SSAImp>
inline typename Meta::size_type SSABase<Meta, SSAImp>::Find(const string_type & pattern,
    size_type & sp, size_type & ep) const {
  size_type i = pattern.size - 1;
  char_type c = pattern[i];

  sp = C(c);
  ep = C(c + 1) - 1;

  //std::cerr << " Looking at [" << sp << ", " << ep << "]\n";
  while (sp <= ep && i >= 1) {
    char_type c = pattern[--i];
    sp = C(c) + Occ(c, sp - 1);
    ep = C(c) + Occ(c, ep) - 1;
    //std::cerr << " Looking at [" << sp << ", " << ep << "]\n";
  }

  //std::cerr << "Found [" << sp << ", " << ep << "]\n";

  if (sp > ep)
    return 0;
  else
    return ep - sp + 1;
}

template <typename Meta, class SSAImp>
bool SSABase<Meta, SSAImp>::Find(const string_type & pattern, State & state) const
{
  size_type i;
  char_type c;

  if (state.mapped == 0) {
    i = pattern.size - 1;
    c = pattern[i];
    state.sps[state.mapped] = C(c);
    state.eps[state.mapped] = C(c + 1) - 1;
    ++state.mapped;
  } else {
    i = pattern.size - state.mapped;
  }

  //std::cerr << " Looking at [" << sp << ", " << ep << "]\n";
  while (state.sps[state.mapped - 1] <= state.eps[state.mapped - 1] && i >= 1) {
    c = pattern[--i];
    state.sps[state.mapped] = C(c) + Occ(c, state.sps[state.mapped - 1] - 1);
    state.eps[state.mapped] = C(c) + Occ(c, state.eps[state.mapped - 1]) - 1;
    //std::cerr << " Looking at [" << sp << ", " << ep << "]\n";
    ++state.mapped;
  }

  if (state.sps[state.mapped - 1] > state.eps[state.mapped - 1])
    --state.mapped;

  if (state.mapped > 0)
    state.count = state.ep() - state.sp() + 1;
  else
    state.count = 0;

  return state.mapped == pattern.size;
}

template <typename Meta, class SSAImp>
inline typename SSABase<Meta, SSAImp>::size_type SSABase<Meta, SSAImp>::Count(const string_type & pattern) const {
  size_type sp, ep;

  return Find(pattern, sp, ep);
}

template <typename Meta, class SSAImp>
inline typename Meta::size_array SSABase<Meta, SSAImp>::Locate(const string_type & pattern) const {
  size_type sp, ep;
  Find(pattern, sp, ep);
  return static_cast<const SSAImp*>(this)->Locate(sp, ep);
}

template <typename Meta>
typename Meta::size_array SSA<Meta>::Locate(size_type sp, size_type ep) const {
  size_array results(ep - sp + 1);

  for (size_type i = 0 ; sp <= ep ; ++sp, ++i) {
    results[i] = BwtToText(sp);
  }

  return results;
}

template <typename Meta>
typename Meta::size_array SSAT<Meta>::Locate(size_type sp, size_type ep) const {
  size_array results(ep - sp + 1);

  for (size_type i = 0 ; sp <= ep ; ++sp, ++i) {
    results[i] = BwtToText(sp);
  }

  return results;
}

template <typename Meta>
typename Meta::size_array SSATS<Meta>::Locate(size_type sp, size_type ep) const {
  size_array results(ep - sp + 1);

  for (size_type i = 0 ; sp <= ep ; ++sp, ++i) {
    results[i] = BwtToText(sp);
  }

  return results;
}

template <typename Meta, class SSAImp>
typename Meta::size_array SSABase<Meta, SSAImp>::Locate(State & state) const {
  return Locate(state, state.count);
}

template <typename Meta, class SSAImp>
typename Meta::size_array SSABase<Meta, SSAImp>::Locate(State & state, size_type count) const {
  return Locate(state, count, 0);
}

template <typename Meta, class SSAImp>
typename Meta::size_array  SSABase<Meta, SSAImp>::Locate(State & state, size_type count, size_type offset) const {
  size_type sp = state.sp() + offset;
  size_type ep = sp + count - 1;
  return static_cast<const SSAImp*>(this)->Locate(sp, ep);
}

template <typename Meta>
typename Meta::string_type SSA<Meta>::Extract(size_type from, size_type length) const {
  string_type slice(length);
  char_type c;
  from = TextToBwt(from + length - 1);
  while (length--) {
    c = BwtOcc(from, from);
    slice[length] = c;
    from += C(c) - 1;
  }

  return slice;
}

template <typename Meta>
typename Meta::string_type SSAT<Meta>::Extract(size_type from, size_type length) const {
  string_type slice(length);
  Extract(from, length, slice.data);
  return slice;
}

template <typename Meta>
typename Meta::string_type SSATS<Meta>::Extract(size_type from, size_type length) const {
  string_type slice(length);
  Extract(from, length, slice.data);
  return slice;
}

template <typename Meta>
void SSA<Meta>::Extract(size_type text_from, size_type length, char_type *to) const
{
  size_type from = TextToBwt(text_from + length - 1);
  char_type c;
  while (length--) {
    c = BwtOcc(from, from);
    to[length] = c;
    from += C(c) - 1;
  }
}

template <typename Meta>
void SSAT<Meta>::Extract(size_type text_from, size_type length, char_type *to) const
{
  text.Extract(text_from, length, to);
}

template <typename Meta>
void SSATS<Meta>::Extract(size_type text_from, size_type length, char_type *to) const
{
  text.Extract(text_from, length, to);
}

template <typename Meta>
typename SSA<Meta>::size_type SSA<Meta>::ExtractLocate(size_type sp, size_type length, char_type *to) const
{
  char_type c;
  size_type text_location = -1u;
  size_type dist = 0;

  while (length--) {
    // Get the text location if possible
    if (text_location == -1u && sampled_bwt[sp]) {
      text_location = bwt_to_text[sampled_bwt.Rank(sp) - 1] + dist;

      // Check whether we are at text[0]
      if (text_location == dist) {
        c = BwtOcc(sp, sp);
        to[length] = c;
        // Fill the rest with Ns
        if (length > 0)
          return -1;
        return text_location;
      }
    } else if (text_location == -1u) {
      ++dist;
    }

    c = BwtOcc(sp, sp);
    to[length] = c;
    sp += C(c) - 1;
  }

  if (text_location == -1u) {
    text_location = BwtToText(sp) + dist;
  }

  return text_location;
}

template <typename Meta>
typename SSAT<Meta>::size_type SSAT<Meta>::ExtractLocate(size_type sp, size_type length, char_type *to) const
{
  size_type text_location = BwtToText(sp);
  if (text_location < length)
      return -1;
  Extract(text_location - length, length, to);
  return text_location;
}

template <typename Meta>
typename SSATS<Meta>::size_type SSATS<Meta>::ExtractLocate(size_type sp, size_type length, char_type *to) const
{
  size_type text_location = BwtToText(sp);
  if (text_location < length)
      return -1;
  Extract(text_location - length, length, to);
  return text_location;
}

template <typename I>
struct SSABuilder;

template <typename Meta>
struct SSABuilder<SSA<Meta> > {
  template <typename Indexer>
  static SSA<Meta> Build(Indexer & indexer, const BWT<Meta> & bwt) {
    SSA<Meta> ssa;
    ssa.Build(bwt);
    return ssa;
  }
};

template <typename Meta>
struct SSABuilder<SSAT<Meta> > {
  template <typename Indexer>
  static SSAT<Meta> Build(Indexer & indexer, const BWT<Meta> & bwt) {
    SSAT<Meta> ssa;
    ssa.Build(bwt, indexer.text());
    return ssa;
  }
};

template <typename Meta>
struct SSABuilder<SSATS<Meta> > {
  template <typename Indexer>
  static SSATS<Meta> Build(Indexer & indexer, const BWT<Meta> & bwt) {
    SSATS<Meta> ssa;
    ssa.Build(bwt, indexer.text());
    return ssa;
  }
};

#endif
