#ifndef _SSA_CUDA_HH
#define _SSA_CUDA_HH 1

#include <iostream>

#include <boost/static_assert.hpp>

#include "config.hh"
#include "util/misc.hh"
#include "util/array_cuda.hh"
#include "util/compact_cuda.hh"
#include "util/timer.hh"
#include "bits_cuda.hh"

#include "ssa_base.hh"
#include "ssa.hh"

template <typename Meta, class SSAImp>
class KernelSSABase {
 public:
  typedef SSABaseSiblings<Meta, SSAImp> SiblingTypes;

  typedef typename Meta::size_type size_type;
  typedef typename Meta::char_type char_type;
  typedef typename Meta::kernel_size_array size_array;
  typedef typename Meta::kernel_string_type string_type;
  typedef typename Meta::kernel_compact_type compact_type;
  typedef typename Meta::kernel_wt_rank_type wt_rank_type;
  static const char_type unknown_char = Meta::unknown_char;
  static const int sigma_size = Meta::sigma_size;
  static const int sample_rate_bwt = Meta::sample_rate_bwt;
  static const int sample_rate_text = Meta::sample_rate_text;
  typedef MatchingResult<KernelSSABase> Result;
  static const int spep_cache_size = Meta::spep_cache_size;
  typedef KernelArray<SpEp<Meta> > spep_cache_type;

  __device__ char_type Bwt(size_type n) const;
  __device__ char_type BwtOcc(size_type n, size_type & r) const;

  __device__ size_type LF(size_type p) const;

  __device__ size_type C(char_type c) const;
  __device__ size_type Occ(char_type c, size_type k) const;
  __device__ size_type OccRank(char_type c, size_type k) const;

  template <typename T>
  __device__ size_type Count(const T & pattern) const;

  template <typename T>
  __device__ size_type Find(const T & pattern, size_type offset, size_type len, size_type & sp, size_type & ep) const;

  template <typename T>
  __device__ void get_cached_spep(const T & pattern, size_type offset, size_type len, size_type & sp, size_type & ep) const;

  size_type size;
  size_type bwt_end_index;
  size_type smaller_than[sigma_size + 1];

  spep_cache_type spep_cache;

  wt_rank_type wt_root;
  wt_rank_type wt_leaves[2];
};

template <typename Meta>
class KernelSSA : public KernelSSABase<Meta, KernelSSA<Meta> > {
 public:
  typedef SSASiblings<Meta> SiblingTypes;

  typedef typename Meta::size_type size_type;
  typedef typename Meta::char_type char_type;
  typedef typename Meta::kernel_size_array size_array;
  typedef typename Meta::kernel_string_type string_type;
  typedef typename Meta::kernel_compact_type compact_type;
  typedef typename Meta::kernel_sampled_bwt_rank_type sampled_bwt_rank_type;

  using KernelSSABase<Meta, KernelSSA<Meta> >::unknown_char;

  using KernelSSABase<Meta, KernelSSA<Meta> >::size;
  using KernelSSABase<Meta, KernelSSA<Meta> >::bwt_end_index;

  using KernelSSABase<Meta, KernelSSA<Meta> >::sample_rate_bwt;
  using KernelSSABase<Meta, KernelSSA<Meta> >::sample_rate_text;

  __device__ size_type BwtToText(size_type n) const;
  __device__ size_type TextToBwt(size_type n) const;

  template <typename A>
  __device__ int CompareToText(size_type text_from, size_type length, A pattern) const;

  size_array text_to_bwt;
  sampled_bwt_rank_type sampled_bwt;
  size_array bwt_to_text;
};

template <typename Meta>
class KernelSSAT : public KernelSSABase<Meta, KernelSSAT<Meta> > {
 public:
  typedef SSATSiblings<Meta> SiblingTypes;

  typedef typename Meta::size_type size_type;
  typedef typename Meta::char_type char_type;
  typedef typename Meta::kernel_size_array size_array;
  typedef typename Meta::kernel_string_type string_type;
  typedef typename Meta::kernel_compact_type compact_type;
  typedef typename Meta::kernel_sampled_bwt_rank_type sampled_bwt_rank_type;

  using KernelSSABase<Meta, KernelSSAT<Meta> >::unknown_char;

  using KernelSSABase<Meta, KernelSSAT<Meta> >::size;
  using KernelSSABase<Meta, KernelSSAT<Meta> >::bwt_end_index;

  using KernelSSABase<Meta, KernelSSAT<Meta> >::sample_rate_bwt;

  __device__ size_type BwtToText(size_type n) const;
  __device__ size_type TextToBwt(size_type n) const;

  template <typename A>
  __device__ int CompareToText(size_type text_from, size_type length, A pattern) const;

  compact_type text;

  sampled_bwt_rank_type sampled_bwt;
  size_array bwt_to_text;
};

template <typename Meta, class CudaSSAImp>
class CudaSSABase {
 public:
  typedef SSABaseSiblings<Meta, CudaSSAImp> SiblingTypes;
  typedef typename SiblingTypes::Host HSSABase;
  typedef typename SiblingTypes::Kernel KSSABase;

  typedef typename Meta::size_type size_type;
  typedef typename Meta::char_type char_type;
  typedef typename Meta::cuda_size_array size_array;
  typedef typename Meta::cuda_string_type string_type;
  typedef typename Meta::cuda_wt_rank_type wt_rank_type;
  typedef CudaArray<SpEp<Meta> > spep_cache_type;

  static const char_type unknown_char = Meta::unknown_char;
  static const int sigma_size = Meta::sigma_size;

  BOOST_STATIC_ASSERT(sigma_size == 4);

  CudaSSABase(const HSSABase & ssab) {
    size = ssab.size;
    bwt_end_index = ssab.bwt_end_index;

    for (int i = 0 ; i < sigma_size + 1; ++i)
      smaller_than[i] = ssab.smaller_than[i];

    wt_root = ssab.wt_root;
    wt_leaves[0] = ssab.wt_leaves[0];
    wt_leaves[1] = ssab.wt_leaves[1];
    spep_cache = ssab.spep_cache;
  }

  size_type size;

 protected:
  void ForKernel(KSSABase & ksb) const {
    ksb.size = size;
    ksb.bwt_end_index = bwt_end_index;

    for (int i = 0 ; i < sigma_size + 1; ++i)
      ksb.smaller_than[i] = smaller_than[i];

    ksb.wt_root = wt_root.ForKernel();
    ksb.wt_leaves[0] = wt_leaves[0].ForKernel();
    ksb.wt_leaves[1] = wt_leaves[1].ForKernel();
    ksb.spep_cache = spep_cache.ForKernel();
  }

  size_type bwt_end_index;

  spep_cache_type spep_cache;

  size_type smaller_than[sigma_size + 1];

  wt_rank_type wt_root;
  wt_rank_type wt_leaves[2];
};

template <typename Meta>
class CudaSSA : public CudaSSABase<Meta, CudaSSA<Meta> > {
 public:
  typedef SSASiblings<Meta> SiblingTypes;
  typedef typename SiblingTypes::Host HSSA;
  typedef typename SiblingTypes::Kernel KSSA;

  typedef typename Meta::size_type size_type;
  typedef typename Meta::char_type char_type;
  typedef typename Meta::cuda_size_array size_array;
  typedef typename Meta::cuda_sampled_bwt_rank_type sampled_bwt_rank_type;

  using CudaSSABase<Meta, CudaSSA<Meta> >::unknown_char;

  using CudaSSABase<Meta, CudaSSA<Meta> >::size;
  using CudaSSABase<Meta, CudaSSA<Meta> >::bwt_end_index;

  CudaSSA(const HSSA & ssa) :
      CudaSSABase<Meta, CudaSSA<Meta> >(ssa),
      text_to_bwt(ssa.text_to_bwt),
      sampled_bwt(ssa.sampled_bwt),
      bwt_to_text(ssa.bwt_to_text)
  {
  }

  KSSA ForKernel() const {
    KSSA ksa;
    CudaSSABase<Meta, CudaSSA<Meta> >::ForKernel(ksa);
    ksa.text_to_bwt = text_to_bwt.ForKernel();
    ksa.sampled_bwt = sampled_bwt.ForKernel();
    ksa.bwt_to_text = bwt_to_text.ForKernel();

    return ksa;
  }

 protected:
  size_array text_to_bwt;
  sampled_bwt_rank_type sampled_bwt;
  size_array bwt_to_text;
};

template <typename Meta>
class CudaSSAT : public CudaSSABase<Meta, CudaSSAT<Meta> > {
 public:
  typedef SSATSiblings<Meta> SiblingTypes;
  typedef typename SiblingTypes::Host HSSA;
  typedef typename SiblingTypes::Kernel KSSA;

  typedef typename Meta::size_type size_type;
  typedef typename Meta::char_type char_type;
  typedef typename Meta::cuda_string_type string_type;
  typedef typename Meta::cuda_size_array size_array;
  typedef typename Meta::cuda_compact_type compact_type;
  typedef typename Meta::cuda_sampled_bwt_rank_type sampled_bwt_rank_type;

  using CudaSSABase<Meta, CudaSSAT<Meta> >::unknown_char;

  using CudaSSABase<Meta, CudaSSAT<Meta> >::size;
  using CudaSSABase<Meta, CudaSSAT<Meta> >::bwt_end_index;

  CudaSSAT(const HSSA & ssa) :
      CudaSSABase<Meta, CudaSSAT<Meta> >(ssa),
      text(ssa.text),
      sampled_bwt(ssa.sampled_bwt),
      bwt_to_text(ssa.bwt_to_text)
  {
  }

  KSSA ForKernel() const {
    KSSA ksa;
    CudaSSABase<Meta, CudaSSAT<Meta> >::ForKernel(ksa);
    ksa.text = text.ForKernel();
    ksa.sampled_bwt = sampled_bwt.ForKernel();
    ksa.bwt_to_text = bwt_to_text.ForKernel();

    return ksa;
  }

 protected:
  compact_type text;

  sampled_bwt_rank_type sampled_bwt;
  size_array bwt_to_text;
};

template <typename Meta, class SSAImp>
template <typename T>
void KernelSSABase<Meta, SSAImp>::get_cached_spep(const T & pattern, size_type offset, size_type len, size_type & sp, size_type & ep) const {
  //get integer representation
  uint ind = 0;
  for(int i = 0; i < spep_cache_size; i++) {
    ind <<= Meta::bits_per_char;
    ind |= pattern[offset + len - i - 1];
  }
  sp = spep_cache[ind].sp;
  ep = spep_cache[ind].ep;
}

template <typename Meta, class SSAImp>
inline typename Meta::char_type KernelSSABase<Meta, SSAImp>::Bwt(size_type n) const {
  size_type r;
  return BwtOcc(n, r);
}

template <typename Meta, class SSAImp>
inline typename Meta::char_type KernelSSABase<Meta, SSAImp>::BwtOcc(size_type n, size_type & r) const {
  // TODO debranching, used by index_to_text
  if (! wt_root[n]) {
    r = wt_root.Rank0(n) - 1;
    if (! wt_leaves[0][r]) {
      r = wt_leaves[0].Rank0(r);
      if (n > bwt_end_index)
        --r;
      else if (n == bwt_end_index)
        r = 0;
      return 0;
    } else {
      r = wt_leaves[0].Rank(r);
      return 1;
    }
  } else {
    r = wt_root.Rank(n) - 1;
    if (! wt_leaves[1][r]) {
      r = wt_leaves[1].Rank0(r);
      return 2;
    } else {
      r = wt_leaves[1].Rank(r);
      return 3;
    }
  }
}

template <typename Meta, class SSAImp>
inline typename Meta::size_type KernelSSABase<Meta, SSAImp>::LF(size_type p) const {
  size_type occ;
  char_type c = BwtOcc(p, occ);
  p = C(c) + occ - 1;
  return p;
}

template <typename Meta>
typename Meta::size_type KernelSSA<Meta>::TextToBwt(size_type n) const {
  size_type p = text_to_bwt[ceildiv(n, sample_rate_text)];

  if (p == bwt_end_index) {
    n = size - n;
  } else {
    // Extra mod to handle exact hit
    n = (sample_rate_text - n % sample_rate_text) % sample_rate_text;
  }

  while (n--) {
    p = LF(p);
  }

  return p;
}

template <typename Meta>
typename Meta::size_type KernelSSA<Meta>::BwtToText(size_type n) const {
  size_type dist = 0;

  while (! sampled_bwt[n]) {
    n = LF(n);
    ++dist;
  }

  return bwt_to_text[sampled_bwt.Rank(n) - 1] + dist;
}

template <typename Meta>
typename Meta::size_type KernelSSAT<Meta>::BwtToText(size_type n) const {
  size_type dist = 0;

  while (! sampled_bwt[n]) {
    n = LF(n);
    ++dist;
  }

  return bwt_to_text[sampled_bwt.Rank(n) - 1] + dist;
}

template <typename Meta>
template <typename A>
__forceinline__ int KernelSSA<Meta>::CompareToText(size_type text_from, size_type length, A pattern) const {
  //TODO compare w/o explicit text
  return 999;
}

template <typename Meta>
template <typename A>
__forceinline__ int KernelSSAT<Meta>::CompareToText(size_type text_from, size_type length, A pattern) const {
  int err = 0;
  for (size_type i = 0; i < length; ++i)
    err += text[text_from + i] != pattern[i];
  return err;
}

template <typename Meta, class SSAImp>
inline typename Meta::size_type KernelSSABase<Meta, SSAImp>::C(char_type c) const {
  return smaller_than[c];
}

template <typename Meta, class SSAImp>
inline typename Meta::size_type KernelSSABase<Meta, SSAImp>::Occ(char_type c, size_type k) const {
  return OccRank(c, k);
}

template <typename Meta, class SSAImp>
typename Meta::size_type KernelSSABase<Meta, SSAImp>::OccRank(char_type c, size_type k) const {
  size_type r = wt_root.RankM(k, c/2) - 1;
  r = wt_leaves[c/2].RankM(r, c & 1);
  r -= (c == 0) && (k >= bwt_end_index);
  return r;
}

template <typename Meta, class SSAImp>
template <typename T>
__forceinline__ typename Meta::size_type KernelSSABase<Meta, SSAImp>::Find(const T & pattern, size_type offset, size_type len, size_type & sp, size_type & ep) const {

  size_type i;
  char_type c;

  if (len < spep_cache_size || spep_cache_size == 0) {
    i = offset + len - 1;
    c = pattern[i];

    sp = C(c);
    ep = C(c + 1) - 1;
  } else {
    get_cached_spep(pattern, offset, len, sp, ep);
    i = offset + len - spep_cache_size;
  }

  while (sp <= ep && i > offset) {
    char_type c = pattern[--i];
    sp = C(c) + Occ(c, sp - 1);
    ep = C(c) + Occ(c, ep) - 1;
  }

  if (sp > ep)
    return 0;
  else
    return ep - sp + 1;
}

template <typename Meta, class SSAImp>
template <typename T>
__forceinline__ typename KernelSSABase<Meta, SSAImp>::size_type KernelSSABase<Meta, SSAImp>::Count(const T & pattern) const {
  size_type sp, ep;

  return Find(pattern, sp, ep);
}

#endif
