#include <gtest/gtest.h>

#include "dna.hh"
#include "ssa.hh"
#include "index.hh"

using namespace bmap;
using dna::Map;

using std::string;

template <typename SSA>
struct Builder
{
};

template <>
struct Builder<SSA<DnaMeta>>
{
  static SSA<DnaMeta> Build(const std::string & dna) {
    return BuildSSA(BuildBWT<DnaMeta>(dna::Map(dna)));
  }
};

template <>
struct Builder<SSAT<DnaMeta>>
{
  static SSAT<DnaMeta> Build(const std::string & dna) {
    return BuildSSAT(BuildBWT<DnaMeta>(dna::Map(dna)), dna::Map(dna));
  }
};

template <>
struct Builder<SSATS<DnaMeta>>
{
  static SSATS<DnaMeta> Build(const std::string & dna) {
    return BuildSSATS(BuildBWT<DnaMeta>(dna::Map(dna)), dna::Map(dna));
  }
};

using testing::Types;

template <typename T>
struct SSATest : testing::Test {
  T Build(const std::string & dna) {
    return Builder<T>::Build(dna);
  }
};

typedef Types<SSA<DnaMeta>, SSAT<DnaMeta>, SSATS<DnaMeta>> SSAs;

TYPED_TEST_CASE(SSATest, SSAs);

TYPED_TEST(SSATest, Create)
{
  this->Build("ACGT");
  this->Build("ACGTACAAGTGTGCGCGAGGAGAGAGAGGTG");
}

template <typename SSA>
::testing::AssertionResult Counts(std::string dna, int multi=0)
{
  while (multi--)
    dna = dna + dna;

  auto ssa = Builder<SSA>::Build(dna);
  auto dnas = Map(dna);

  unsigned count[4] = {0u, 0u, 0u, 0u};
  DnaString bases[4] = {dna::Map("a"), dna::Map("c"), dna::Map("g"), dna::Map("t")};

  for (auto i = 0u; i < dnas.size; ++i)
    if (dnas[i] < 4)
      ++count[dnas[i]];

  for (auto b = dna::A; b <= dna::T; ++b)
    if (count[b] != ssa.Count(bases[b]))
      return ::testing::AssertionFailure()
        << "Count " << dna::Unmap(bases[b]) << " expected " << count[b] << " != " << ssa.Count(bases[b]);

  return ::testing::AssertionSuccess();
}

TYPED_TEST(SSATest, Count1)
{
  EXPECT_TRUE(Counts<TypeParam>(""));
  EXPECT_TRUE(Counts<TypeParam>("A"));
  EXPECT_TRUE(Counts<TypeParam>("ACGT"));
  EXPECT_TRUE(Counts<TypeParam>("AACCGGTT"));
  EXPECT_TRUE(Counts<TypeParam>("ACGTACGT"));
  EXPECT_TRUE(Counts<TypeParam>("AAAAAAAAAAAAAA"));
  EXPECT_TRUE(Counts<TypeParam>("CCCCCCCCCCCCCC"));
  EXPECT_TRUE(Counts<TypeParam>("GGGGGGGGGGGGGG"));
  EXPECT_TRUE(Counts<TypeParam>("TTTTTTTTTTTTTT"));
}

TYPED_TEST(SSATest, Count1Big)
{
  EXPECT_TRUE(Counts<TypeParam>("AAAA", 12));
  EXPECT_TRUE(Counts<TypeParam>("CCCC", 12));
  EXPECT_TRUE(Counts<TypeParam>("ACGT", 12));
  EXPECT_TRUE(Counts<TypeParam>("TCGA", 12));
  EXPECT_TRUE(Counts<TypeParam>("CGAT", 12));
  EXPECT_TRUE(Counts<TypeParam>("CGAACGAT", 10));
}


TYPED_TEST(SSATest, Count1Random)
{
  DnaString d{80};
  for (auto i = 0u ; i < 10000u ; ++i) {
    dna::Randomize(d);
    EXPECT_TRUE(Counts<TypeParam>(dna::Unmap(d))) << "for '" << dna::Unmap(d) << "'";
  }
}

TYPED_TEST(SSATest, CountMore)
{
  EXPECT_EQ(3, this->Build("AAAA").Count(Map("AA")));
  EXPECT_EQ(2, this->Build("AAAA").Count(Map("AAA")));
  EXPECT_EQ(1, this->Build("AAAA").Count(Map("AAAA")));
  EXPECT_EQ(0, this->Build("AAAA").Count(Map("AAAAA")));

  EXPECT_EQ(3, this->Build("CCCC").Count(Map("CC")));
  EXPECT_EQ(2, this->Build("CCCC").Count(Map("CCC")));
  EXPECT_EQ(1, this->Build("CCCC").Count(Map("CCCC")));
  EXPECT_EQ(0, this->Build("CCCC").Count(Map("CCCCC")));

  EXPECT_EQ(1, this->Build("AC").Count(Map("AC")));
  EXPECT_EQ(0, this->Build("AC").Count(Map("CA")));

  EXPECT_EQ(1, this->Build("ACGT").Count(Map("ACGT")));
  EXPECT_EQ(0, this->Build("ACGT").Count(Map("TACG")));
  EXPECT_EQ(0, this->Build("ACGT").Count(Map("GTAC")));
  EXPECT_EQ(0, this->Build("ACGT").Count(Map("CGTA")));
}


template <typename SSA>
::testing::AssertionResult Extracts(std::string dna, int multi=0)
{
  while (multi--)
    dna = dna + dna;

  auto ssa = Builder<SSA>::Build(dna);
  auto dnas = Map(dna);

  for (auto from = 0u; from < dnas.size; ++from) {
    for (auto len = 1u; from + len <= dnas.size ; ++len) {
      auto exp = dnas.Slice(from, len);
      auto got = ssa.Extract(from, len);
      if (exp != got)
        return ::testing::AssertionFailure()
          << "Extract [" << from << ", " << from + len - 1 << "] expected '"
          << dna::Unmap(exp)  << "' != '" << dna::Unmap(got) << "'";
    }
  }

  return ::testing::AssertionSuccess();
}

TYPED_TEST(SSATest, Extract)
{
  EXPECT_TRUE(Extracts<TypeParam>(""));
  EXPECT_TRUE(Extracts<TypeParam>("ACGT"));
  EXPECT_TRUE(Extracts<TypeParam>("AAAA"));
  EXPECT_TRUE(Extracts<TypeParam>("CCCC"));
  EXPECT_TRUE(Extracts<TypeParam>("GGGG"));
  EXPECT_TRUE(Extracts<TypeParam>("TTTT"));
  EXPECT_TRUE(Extracts<TypeParam>("AAA", 5));
  EXPECT_TRUE(Extracts<TypeParam>("CCC", 5));
  EXPECT_TRUE(Extracts<TypeParam>("GGG", 5));
  EXPECT_TRUE(Extracts<TypeParam>("TTT", 5));
  EXPECT_TRUE(Extracts<TypeParam>("ACG", 5));
  EXPECT_TRUE(Extracts<TypeParam>("TGC", 5));
  EXPECT_TRUE(Extracts<TypeParam>("TGCAC", 5));
}

template <typename SSA>
::testing::AssertionResult Locates(std::string dna, int multi=0)
{
  while (multi--)
    dna = dna + dna;

  auto ssa = Builder<SSA>::Build(dna);
  auto dnas = Map(dna);

  for (auto from = 0u; from < dnas.size; ++from) {
    for (auto len = 1u; len < 128 && from + len <= dnas.size ; ++len) {
      auto exp = dnas.Slice(from, len);
      auto locs = ssa.Locate(exp);
      for (auto i = 0u; i < locs.size; ++i) {
        if (locs[i] + len > dnas.size)
          return ::testing::AssertionFailure()
            << "Locating '" << dna::Unmap(exp) << "' in '" << dna::Unmap(dnas)
            << "' out of bounds location " << i << ": " << locs[i];
        auto slice = dnas.Slice(locs[i], len);
        if (slice != exp)
          return ::testing::AssertionFailure()
            << "Locating '" << dna::Unmap(exp) << "' in '" << dna::Unmap(dnas)
            << "' bad location " << i << ": " << locs[i] << " '" << dna::Unmap(slice) << "'";
      }
      if (locs.size == 0)
        return ::testing::AssertionFailure()
          << "Locating '" << dna::Unmap(exp) << "' in '" << dna::Unmap(dnas) << "' failed";
    }
  }

  return ::testing::AssertionSuccess();
}

TYPED_TEST(SSATest, Locate)
{
  EXPECT_TRUE(Locates<TypeParam>(""));
  EXPECT_TRUE(Locates<TypeParam>("ACGT"));
  EXPECT_TRUE(Locates<TypeParam>("AAAA"));
  EXPECT_TRUE(Locates<TypeParam>("CCCC"));
  EXPECT_TRUE(Locates<TypeParam>("ACGTACGTACACAGT"));
}

TYPED_TEST(SSATest, LocateRandom)
{
  DnaString d{80};
  for (auto i = 0u ; i < 100u ; ++i) {
    dna::Randomize(d);
    EXPECT_TRUE(Locates<TypeParam>(dna::Unmap(d))) << "for '" << dna::Unmap(d) << "'";
  }
}

template <typename SSA>
::testing::AssertionResult ExtractLocates(std::string dna, int multi=0)
{
  while (multi--)
    dna = dna + dna;

  auto ssa = Builder<SSA>::Build(dna);
  auto dnas = Map(dna);
  typedef typename SSA::State State;
  typedef typename SSA::string_type string_type;

  for (auto from = 1u; from < dnas.size; ++from) {
    for (auto len = 1u; len < 128 && from + len <= dnas.size; ++len) {
      State state;
      auto pattern = dnas.Slice(from, len);
      auto expected = dnas.Slice(0, from);
      string_type extracted(expected.size);
      ssa.Find(pattern, state);
      auto res = state.result();
      for (auto i = 0u; i < res.count; ++i) {
        if (ssa.BwtToText(res.sp + i) != from)
          continue;
        auto text_location = ssa.ExtractLocate(res.sp + i, from, extracted.data);
        if (text_location != from)
          return ::testing::AssertionFailure()
            << "ExtractLocating '" << dna::Unmap(pattern) << "' in '" << dna::Unmap(dnas)
            << "' bad text location " << text_location << " != " << from;
        if (extracted != expected)
          return ::testing::AssertionFailure()
            << "ExtractLocating '" << dna::Unmap(pattern) << "' in '" << dna::Unmap(dnas)
            << "' bad extract '" << dna::Unmap(extracted) << "' != '" << dna::Unmap(expected) << "'";
      }
    }
  }

  return ::testing::AssertionSuccess();
}

TYPED_TEST(SSATest, ExtractLocate)
{
  EXPECT_TRUE(ExtractLocates<TypeParam>(""));
  EXPECT_TRUE(ExtractLocates<TypeParam>("ACGT"));
  EXPECT_TRUE(ExtractLocates<TypeParam>("AAAA"));
  EXPECT_TRUE(ExtractLocates<TypeParam>("CCCC"));
  EXPECT_TRUE(ExtractLocates<TypeParam>("ACGTACGTACACAGT"));
  EXPECT_TRUE(ExtractLocates<TypeParam>("ACGTACGTACACAGT", 4));
}
