#include <fstream>
#include <string>
#include <tr1/unordered_map>
#include <vector>
#include <functional>
#include <algorithm>
#include <limits>

#include "util/serialization.hh"
#include "util/timer.hh"
#include "bwt.hh"
#include "ssa.hh"
#include "index.hh"

using namespace bmap;

template <int Size>
struct ShortDna {
  static const uint64_t MASK = (1ULL << Size * 2) - 1;
  uint64_t dna;
  ShortDna() : dna(0) {}

  void add(int a) {
    dna <<= 2;
    dna |= a;
    dna &= MASK;
  }

  DnaString decode() const {
    DnaString ds(Size);
    for (int i = 0; i < Size; ++i) {
      ds[i] = (dna >> (2 * (Size - i - 1))) & 3;
    }

    return ds;
  }
};

struct Histogram {
  uint64_t total;
  uint64_t count;
  std::vector<std::pair<std::string, uint64_t>> ranges;
  std::vector<uint64_t> counts;

  Histogram(const std::vector<std::pair<std::string, uint64_t>> & r) {
    total = 0;
    count = 0;
    ranges = r;
    counts.assign(ranges.size(), 0);
  }

  Histogram() {
    total = 0;
    count = 0;
    ranges = {
      {"1", 1},
      {"2", 2},
      {"3", 3},
      {"4", 4},
      {"5", 5},
      {"-10", 10},
      {"-50", 50},
      {"-100", 100},
      {"-500", 500},
      {"-1000", 1000},
      {"-10000", 10000},
      {"-100000", 100000},
      {"100000+", std::numeric_limits<uint64_t>::max()},
    };
    counts.assign(ranges.size(), 0);
  }

  void add(uint64_t n) {
    total += n;
    ++count;
    for (uint i = 0; i < ranges.size(); ++i) {
      if (n <= ranges[i].second) {
        ++counts[i];
        break;
      }
    }
  }

  void print() {
    std::cout << "count: " << count << " total: " << total << " avg: " << (double(total) / double(count)) << "\n";
    for (uint i = 0; i < ranges.size(); ++i)
      std::cout << ranges[i].first << " " << counts[i] << " " << double(counts[i]) / count << "\n";
  }
};

void n_stats(DnaString & d) {
  Histogram hist{std::vector<std::pair<std::string, uint64_t>>{
    {"1", 1},
    {"-5", 5},
    {"-10", 10},
    {"-50", 50},
    {"-100", 100},
    {"-500", 100},
    {"-1000", 1000},
    {"-10000", 10000},
    {"-100000", 100000},
    {"-1000000", 1000000},
    {"1000000+", std::numeric_limits<uint64_t>::max()},
  }};

  uint64_t run = 0;
  for (size_t i = 0; i < d.size; ++i) {
    if (d[i] == dna::G) {
      ++run;
    } else if (run != 0) {
      hist.add(run);
      run = 0;
    }
  }
  if (run != 0)
    hist.add(run);
  hist.print();
}

template <int Size>
void stats(DnaString & dna) {
  const uint64_t N = 1<<28;
  std::vector<std::pair<int, uint64_t>> res;
  res.reserve(10 * 1000 * 1000);

  uint64_t sum = 0;
  uint64_t count = 0;
  Histogram hist;

  for (uint suffix = 0; suffix < 16; ++suffix) {
#if 1
    std::tr1::unordered_map<uint64_t, int> stats(N);
    stats.max_load_factor(1000);
#else
    std::tr1::unordered_map<uint64_t, int> stats;
#endif
    ShortDna<Size> d;
    for (int i = 0; i < Size - 1; ++i)
      d.add(dna[i]);

    for (size_t i = Size - 1; i < dna.size; ++i) {
      d.add(dna[i]);
      if ((d.dna & 15) == suffix)
        ++stats[d.dna];
      if (i % 10000000 == 0)
        std::cerr << "Done " << suffix + 1 << " / 16 " << i << " / " << dna.size
          << " buckets: " << stats.bucket_count() << " factor: " << stats.load_factor() << " / " << stats.max_load_factor() << "\n";
      //if (i % 100000000 == 0)
      //break;
    }

    std::cerr << "Counting\n";
    for (auto i = stats.begin(); i != stats.end(); ++i) {
      res.push_back(std::make_pair(i->second, i->first));
      ++count;
      sum += i->second;
      hist.add(i->second);
      if (res.size() == 9900000) {
        std::partial_sort(res.begin(), res.begin() + 20, res.end(), std::greater<std::pair<int, uint64_t>>());
        res.resize(20);
      }
    }
    std::partial_sort(res.begin(), res.begin() + 20, res.end(), std::greater<std::pair<int, uint64_t>>());
    res.resize(20);
    std::cerr << "Counting DONE\n";
  }
  dna.Clear();


  std::cout << count << " different substrings of size " << Size << "\n";
  std::cout << " avg " << (double(sum) / count) << " occ\n";

  for (int i = 0; i < 20; ++i) {
    ShortDna<Size> d;
    d.dna = res[i].second;
    std::cout << "Most " << res[i].first << " of " << dna::Unmap(d.decode()) << "\n";
  }
  hist.print();
}

int main(int argc, char * argv[]) {
  srand(31337);
  Timer timer;

  if (argc != 3 && argc != 2) {
    std::cerr << "Usage: " << argv[0] << " fasta-file len\n";
    return 1;
  }

  std::string fasta_f = argv[1];
  std::ifstream fasta_s{fasta_f};

  int len = 0;

  if (argc == 3)
    len = atoi(argv[2]);

  if (! fasta_s.good()) {
    std::cerr << "Bad file " << fasta_f << "\n";
    return EXIT_FAILURE;
  }

  dna::FastaReader fasta_reader{fasta_s};
  DnaString dna_string;
  fasta_reader.read_dna(dna_string);
  //dna::RandomizeN(dna_string);

  if (len == 0) {
    n_stats(dna_string);
    return 0;
  }

  dna_string = dna::Strip(dna_string, dna::N);

  std::cerr << "Read dna\n";
  switch (len) {
  case 0:
    n_stats(dna_string);
    break;
  case 10:
    stats<10>(dna_string);
    break;
  case 11:
    stats<11>(dna_string);
    break;
  case 12:
    stats<12>(dna_string);
    break;
  case 13:
    stats<13>(dna_string);
    break;
  case 14:
    stats<14>(dna_string);
    break;
  case 15:
    stats<15>(dna_string);
    break;
  case 16:
    stats<16>(dna_string);
    break;
  case 17:
    stats<17>(dna_string);
    break;
  case 18:
    stats<18>(dna_string);
    break;
  case 19:
    stats<19>(dna_string);
    break;
  case 20:
    stats<20>(dna_string);
    break;
  case 21:
    stats<21>(dna_string);
    break;
  case 22:
    stats<22>(dna_string);
    break;
  case 23:
    stats<23>(dna_string);
    break;
  case 24:
    stats<24>(dna_string);
    break;
  case 25:
    stats<25>(dna_string);
    break;
  case 30:
    stats<30>(dna_string);
    break;
  default:
    std::cerr << "Bad len\n";
    return EXIT_FAILURE;
  }
}
