#ifndef _UTIL_ARRAY_HH_
#define _UTIL_ARRAY_HH_

#include <stdint.h>
#include <assert.h>

#include <algorithm>
#include <vector>
#include <string>

#include <boost/type_traits.hpp>
#include <boost/static_assert.hpp>
#include <boost/serialization/serialization.hpp>
#include <boost/serialization/split_member.hpp>
#include <boost/serialization/array.hpp>

#include <cuda_runtime.h>

#include "config.hh"

#include "util/array_base.hh"

template <typename T>
struct SArray {
  typedef ArraySiblings<T> SiblingTypes;

  typedef uint32_t size_type;
  typedef T value_type;

  size_type size;
  T * data;

  SArray() : size(0), data(NULL) {}
  SArray(size_type n) : size(n), data(new T[n]) {}
  SArray(size_type n, T t);
  SArray(size_type n, T * d);
  SArray(std::vector<T> v);
  SArray(std::basic_string<T> v);

  SArray(const SArray<T> & other);
  SArray<T> & operator=(const SArray<T> & other);

  SArray(const CudaArray<T> & other);
  SArray<T> & operator=(const CudaArray<T> & other);

#ifdef __CXX0X__
  SArray(SArray<T> && other);
  SArray<T> & operator=(SArray<T> && other);
#endif

  ~SArray() { delete[] data; }

  void Clear();
  void Resize(size_type n);

  T & operator[](size_type n) { assert(n<size); return data[n]; }
  const T & operator[](size_type n) const { assert(n<size); return data[n]; }

  bool operator==(SArray<T> & other) const;
  bool operator!=(SArray<T> & other) const;

  SArray<T> Slice(size_type from, size_type len) const;
  template <typename A>
  void Slice(size_type from, size_type len, A a) const;

 private:
  friend class boost::serialization::access;
  template <typename Archive>
  void save(Archive & ar, const unsigned int version) const;
  template <typename Archive>
  void load(Archive & ar, const unsigned int version);
  BOOST_SERIALIZATION_SPLIT_MEMBER()
};

template <typename T>
template <typename Archive>
void SArray<T>::save(Archive & ar, const unsigned int version) const {
  ar & size;
  ar & boost::serialization::make_array(data, size);
}

template <typename T>
template <typename Archive>
void SArray<T>::load(Archive & ar, const unsigned int version) {
  Clear();
  ar & size;
  data = new T[size];
  ar & boost::serialization::make_array(data, size);
}

template <typename T>
SArray<T>::SArray(size_type n, T t) : size(n), data(new T[size]) {
  for (size_type i = 0 ; i < size ; ++i)
    data[i] = t;
}

template <typename T>
SArray<T>::SArray(size_type s, T * d) : size(s), data(new T[size]) {
  std::copy(d, d + size, data);
}

template <typename T>
SArray<T>::SArray(std::vector<T> v) : size(v.size()), data(new T[size]) {
  std::copy(v.begin(), v.end(), data);
}

template <typename T>
SArray<T>::SArray(std::basic_string<T> s) : size(s.size()), data(new T[size]) {
  std::copy(s.begin(), s.end(), data);
}

template <typename T>
SArray<T>::SArray(const SArray<T> & other) : size(other.size), data(new T[size]) {
  std::copy(other.data, other.data + size, data);
}

template <typename T>
SArray<T> & SArray<T>::operator=(const SArray<T> & other) {
  if (this == &other)
    return *this;

  if (size != other.size) {
    delete[] data;
    size = other.size;
    data = new T[size];
  }

  std::copy(other.data, other.data + size, data);

  return *this;
}

template <typename T>
SArray<T>::SArray(const CudaArray<T> & ca) : size(ca.size), data(new T[size]) {
  BOOST_STATIC_ASSERT(boost::has_trivial_copy<T>::value);
  cudaMemcpy(data, ca.data, sizeof(T) * size, cudaMemcpyDeviceToHost);
}

template <typename T>
SArray<T> & SArray<T>::operator=(const CudaArray<T> & ca) {
  BOOST_STATIC_ASSERT(boost::has_trivial_copy<T>::value);

  if (size != ca.size) {
    delete[] data;
    size = ca.size;
    data = new T[size];
  }

  cudaMemcpy(data, ca.data, sizeof(T) * size, cudaMemcpyDeviceToHost);

  return *this;
}

template <typename T>
void SArray<T>::Clear() {
  delete[] data;
  data = NULL;
  size = 0;
}

template <typename T>
void SArray<T>::Resize(size_type n) {
  T *old_data = data;
  data = new T[n];
  std::copy(old_data, old_data + std::min(size, n), data);
  delete[] old_data;
  size = n;
}

#ifdef __CXX0X__
template <typename T>
SArray<T>::SArray(SArray<T> && other) : size(other.size), data(other.data) {
  other.data = NULL;
  other.size = 0;
}

template <typename T>
SArray<T> & SArray<T>::operator=(SArray<T> && other) {
  delete[] data;

  data = other.data;
  size = other.size;
  other.data = NULL;
  other.size = 0;

  return *this;
}
#endif

template <typename T>
SArray<T> SArray<T>::Slice(size_type from, size_type len) const {
  assert(from + len <= size);

  SArray<T> slice(len);
  std::copy(data + from, data + from + len, slice.data);
  return slice;
}

template <typename T>
template <typename A>
void SArray<T>::Slice(size_type from, size_type len, A a) const {
  assert(from + len <= size);
  for (uint i = 0; i < len; ++i)
    a[i] = data[from + i];
}

template <typename T>
bool SArray<T>::operator==(SArray<T> & other) const {
  if (size != other.size)
    return false;

  return memcmp(data, other.data, size) == 0;
}

template <typename T>
bool SArray<T>::operator!=(SArray<T> & other) const {
  return !(*this == other);
}

std::ostream & operator<<(std::ostream & o, const SArray<unsigned char> & sa) {
  o.write((const char*)sa.data, sa.size);

  return o;
}

//#define MAGIC_INDEX
//#define MAGIC_ROW

#ifdef MAGIC_INDEX
template <typename Array2D>
struct Array2DRow {
  typedef typename Array2D::value_type value_type;
  static const int BATCH = Array2D::BATCH;
#ifdef MAGIC_ROW
  value_type * data;
  int size_;
#else
  Array2D & array;
  int row;
#endif

  __host__ __device__ Array2DRow(const Array2D & a, int r) :
#ifdef MAGIC_ROW
    data(const_cast<value_type*>(a.data.data + (r / BATCH) * BATCH * a.cols + r % BATCH))
    ,size_(a.cols)
#else
    array(const_cast<Array2D &>(a)), row(r)
#endif
  {}

  __host__ __device__ value_type & operator[](int n) {
#ifdef MAGIC_ROW
    return data[n * BATCH];
#else
    return array(row, n);
#endif
  }

  __host__ __device__ value_type operator[](int n) const {
#ifdef MAGIC_ROW
    return data[n * BATCH];
#else
    return array(row, n);
#endif
  }

  Array2DRow<Array2D> & operator=(const SArray<value_type> & a) {
    for (uint i = 0; i < a.size; ++i)
      (*this)[i] = a[i];
    return *this;
  }

  __host__ __device__ size_t size() const {
#ifdef MAGIC_ROW
    return size_;
#else
    return array.cols;
#endif
  }
};

#else

template <typename Array2D>
struct Array2DRow {
  typedef typename Array2D::value_type value_type;
  value_type * data;
  int size_;

  __host__ __device__ Array2DRow(const Array2D & a, int r) :
    data(const_cast<value_type *>(a.data.data + r * a.cols)), size_(a.cols) {}

  __host__ __device__ value_type & operator[](int n) {
    return data[n];
  }

  __host__ __device__ value_type operator[](int n) const {
    return data[n];
  }

  Array2DRow<Array2D> & operator=(const SArray<value_type> & a) {
#if 1
    //memcpy(data, a.data, a.size);
    std::copy(a.data, a.data + a.size, data);
#else
    for (uint i = 0; i < a.size; ++i)
      data[i] = a[i];
#endif
    return *this;
  }

  __host__ __device__ size_t size() const {
    return size_;
  }
};

#endif

template <typename T, int B>
struct Array2D {
  const static int BATCH = B;
  typedef T value_type;
  typedef Array2DSiblings<T, BATCH> SiblingTypes;
  typedef typename SiblingTypes::Host H2D;
  typedef Array2DRow<H2D> Row;

  int rows;
  int cols;
  SArray<T> data;

  Array2D(int r, int c) : rows(r), cols(c), data(r * c) {}

  T & operator()(int row, int col) {
#ifdef MAGIC_INDEX
    return data[(row / BATCH) * BATCH * cols + row % BATCH + col * BATCH];
#else
    return data[row * cols + col];
#endif
  }

  void Resize(size_t n) {
    rows = n;
    data.Resize(rows * cols);
  }

  size_t size() const {
    return rows;
  }

  T operator()(int row, int col) const {
    return const_cast<H2D &>(*this)(row, col);
  }

  Row operator[](int row) const {
    return Row(*this, row);
  }
};

#endif // _UTIL_ARRAY_HH_
