#ifndef _UTIL_ARRAY_CUDA_HH_
#define _UTIL_ARRAY_CUDA_HH_

#include <boost/type_traits.hpp>
#include <boost/static_assert.hpp>

#include "config.hh"

#include "util/siblings.hh"
#include "util/array_base.hh"
#include "util/array.hh"

template <typename T>
struct KernelArray {
  typedef ArraySiblings<T> SiblingTypes;

  T *data;
  typedef uint32_t size_type;
  __device__ T operator[](size_type n) const { return data[n]; }
  __device__ T & operator[](size_type n) { return data[n]; }
};

template <typename T>
struct KernelSizedArray {
  typedef ArraySiblings<T> SiblingTypes;

  T *data;
  typedef uint32_t size_type;
  size_type size;
  typedef T value_type;
  __device__ T operator[](size_type n) const { return data[n]; }
  __device__ T & operator[](size_type n) { return data[n]; }
};


template <typename T>
struct CudaArray {
  typedef ArraySiblings<T> SiblingTypes;

  typedef uint32_t size_type;
  typedef T value_type;

  typedef typename SiblingTypes::Kernel KernelA;
  typedef typename SiblingTypes::KernelSized KernelSizedA;

  size_type size;
  T * data;

  CudaArray() : size(0), data(NULL) {}

  CudaArray(size_type n) : size(n) {
    //causes weird incomplete type errors
    //BOOST_STATIC_ASSERT(boost::has_trivial_constructor<T>::value);

    cudaMalloc(&data, sizeof(T) * n);
  }

  CudaArray(const SArray<T> & sa);
  CudaArray(std::vector<T> v);
  CudaArray(std::basic_string<T> v);

  CudaArray(const CudaArray<T> & other);
#ifdef __CXX0X__
  CudaArray(CudaArray<T> && other);
#endif

  CudaArray<T> & operator=(const CudaArray<T> & other);
#ifdef __CXX0X__
  CudaArray<T> & operator=(CudaArray<T> && other);
#endif

  ~CudaArray() { cudaFree(data); }

  KernelA ForKernel() const { KernelA ka; ka.data = data; return ka; }
  KernelSizedA ForKernelSized() const { KernelSizedA ka; ka.size = size; ka.data = data; return ka; }

  void Clear();
  void Resize(size_type n);

  T operator[](size_type n) const {
    BOOST_STATIC_ASSERT(boost::has_trivial_copy<T>::value);
    T a;
    cudaMemcpy(&a, data + n, sizeof(T), cudaMemcpyDeviceToHost);
    return a;
  }

  void Set(size_type n, T a) const {
    BOOST_STATIC_ASSERT(boost::has_trivial_copy<T>::value);
    cudaMemcpy(data + n, &a, sizeof(T), cudaMemcpyHostToDevice);
  }

  CudaArray<T> Slice(size_type from, size_type len) const;
};

template <typename T>
CudaArray<T>::CudaArray(const SArray<T> & sa) : size(sa.size) {
  BOOST_STATIC_ASSERT(boost::has_trivial_copy<T>::value);
  cudaMalloc(&data, sizeof(T) * size);
  cudaMemcpy(data, sa.data, sizeof(T) * size, cudaMemcpyHostToDevice);
}

template <typename T>
CudaArray<T>::CudaArray(std::basic_string<T> s) : size(s.size()) {
  BOOST_STATIC_ASSERT(boost::has_trivial_copy<T>::value);
  cudaMalloc(&data, sizeof(T) * size);
  cudaMemcpy(data, s.c_str(), sizeof(T) * size, cudaMemcpyHostToDevice);
}

template <typename T>
CudaArray<T>::CudaArray(const CudaArray<T> & other) : size(other.size) {
  BOOST_STATIC_ASSERT(boost::has_trivial_copy<T>::value);
  cudaMalloc(&data, sizeof(T) * size);
  cudaMemcpy(data, other.data, sizeof(T) * size, cudaMemcpyDeviceToDevice);
}

#ifdef __CXX0X__
template <typename T>
CudaArray<T>::CudaArray(CudaArray<T> && other) : size(other.size), data(other.data) {
  other.data = NULL;
  other.size = 0;
}
#endif

template <typename T>
void CudaArray<T>::Clear() {
  cudaFree(data);
  data = NULL;
  size = 0;
}

template <typename T>
void CudaArray<T>::Resize(size_type n) {
  BOOST_STATIC_ASSERT(boost::has_trivial_copy<T>::value);
  T *old_data = data;
  cudaMalloc(&data, sizeof(T) * n);
  cudaMemcpy(data, old_data, sizeof(T) * std::min(size, n), cudaMemcpyDeviceToDevice);
  cudaFree(old_data);
  size = n;
}

template <typename T>
CudaArray<T> & CudaArray<T>::operator=(const CudaArray<T> & other) {
  BOOST_STATIC_ASSERT(boost::has_trivial_copy<T>::value);

  if (this == &other)
    return *this;

  if (size != other.size) {
    cudaFree(data);
    size = other.size;
    cudaMalloc(&data, sizeof(T) * size);
  }

  cudaMemcpy(data, other.data, sizeof(T) * size, cudaMemcpyDeviceToDevice);

  return *this;
}

#ifdef __CXX0X__
template <typename T>
CudaArray<T> & CudaArray<T>::operator=(CudaArray<T> && other) {
  cudaFree(data);

  data = other.data;
  size = other.size;
  other.data = NULL;
  other.size = 0;

  return *this;
}
#endif

template <typename T>
CudaArray<T> CudaArray<T>::Slice(size_type from, size_type len) const {
  BOOST_STATIC_ASSERT(boost::has_trivial_copy<T>::value);
  CudaArray<T> slice(len);
  cudaMemcpy(slice.data, data + from, sizeof(T) * len, cudaMemcpyDeviceToDevice);
  return slice;
}

template <typename T, int B>
struct KernelArray2D {
  const static int BATCH = B;
  typedef T value_type;
  typedef Array2DSiblings<T, BATCH> SiblingTypes;
  typedef typename SiblingTypes::Kernel K2D;
  typedef Array2DRow<K2D> Row;

  int rows;
  int cols;
  KernelArray<T> data;

  __device__ T & operator()(int row, int col) {
#ifdef MAGIC_INDEX
    return data[(row / BATCH) * BATCH * cols + row % BATCH + col * BATCH];
#else
    return data[row * cols + col];
#endif
  }

  __device__ T operator()(int row, int col) const {
    return const_cast<K2D &>(*this)(row, col);
  }

  __device__ Row operator[](int row) {
    return Row(*this, row);
  }

  __device__ const Row operator[](int row) const {
    return Row(*this, row);
  }
};

template <typename T, int B>
struct CudaArray2D {
  const static int BATCH = B;
  typedef T value_type;
  typedef Array2DSiblings<T, BATCH> SiblingTypes;
  typedef typename SiblingTypes::Host H2D;
  typedef typename SiblingTypes::Kernel K2D;
  typedef typename SiblingTypes::Cuda C2D;
  typedef Array2DRow<C2D> Row;

  int rows;
  int cols;
  CudaArray<T> data;

  CudaArray2D(int r, int c) : rows(r), cols(c), data(r * c) {}
  CudaArray2D(const H2D & h2d) : rows(h2d.rows), cols(h2d.cols), data(h2d.data) {}

  T & operator()(int row, int col) {
#ifdef MAGIC_INDEX
    return data[(row / BATCH) * BATCH * cols + row % BATCH + col * BATCH];
#else
    return data[row * cols + col];
#endif
  }

  T operator()(int row, int col) const {
    return const_cast<C2D &>(*this)(row, col);
  }

  size_t size() const {
    return rows;
  }

  void Resize(size_t n) {
    rows = n;
    data.Resize(rows * cols);
  }

  K2D ForKernel() const {
    K2D k2d;
    k2d.rows = rows;
    k2d.cols = cols;
    k2d.data = data.ForKernel();
    return k2d;
  }
};

#endif // _UTIL_ARRAY_CUDA_HH_
