#ifndef _UTIL_COMPACT_HH
#define _UTIL_COMPACT_HH

#include <stdint.h>

#include <boost/serialization/serialization.hpp>

#include "config.hh"
#include "util/misc.hh"
#include "util/array.hh"
#include "util/compact_base.hh"

template <typename Base, int ElementSize, typename Element>
class CompactArray {
 public:
  typedef CompactArraySiblings<Base, ElementSize, Element> SiblingTypes;

  CompactArray() : size(0) {}
  CompactArray(size_t s);
  CompactArray(const SArray<Element> & array);

#ifdef __CXXOX__
  CompactArray<Base, ElementSize, Element> & operator==(CompactArray<Base, ElementSize, Element> && other);
#endif

  const static size_t BaseSize = sizeof(Base) * 8 / ElementSize;
  const static Base ElemMask = (1ULL << ElementSize) - 1;

  Element Get(size_t n) const;
  void Set(size_t n, Base el);

  void Extract(size_t from, size_t length, Element * array) const;
  void Extract(size_t from, size_t length, SArray<Element> & array) const;

  Element operator[](size_t n) const;

  size_t size;
  SArray<Base> data;

 private:
  friend class boost::serialization::access;
  template<class Archive>
  void serialize(Archive & ar, const unsigned int version);
};

template <typename Base, int ElementSize, typename Element>
CompactArray<Base, ElementSize, Element>::CompactArray(size_t s) :
  size(s), data(ceildiv(size, BaseSize))
{
}

template <typename Base, int ElementSize, typename Element>
CompactArray<Base, ElementSize, Element>::CompactArray(const SArray<Element> & array) :
  size(array.size), data(ceildiv(size, BaseSize))
{
  for (size_t i = 0; i < size; ++i)
    Set(i, array[i]);
}

#ifdef __CXXOX__
template <typename Base, int ElementSize, typename Element>
CompactArray<Base, ElementSize, Element> &
CompactArray<Base, ElementSize, Element>::operator==(CompactArray<Base, ElementSize, Element> && other)
{
  size = other.size;
  other.size = 0;
  data = std::move(other.data);
}
#endif

template <typename Base, int ElementSize, typename Element>
template <typename Archive>
void CompactArray<Base, ElementSize, Element>::serialize(Archive & ar, const unsigned int version) {
  ar & size;
  ar & data;
}

template <typename Base, int ElementSize, typename Element>
Element CompactArray<Base, ElementSize, Element>::Get(size_t n) const {
  return (data[n / BaseSize] >> (ElementSize * (n % BaseSize))) & ElemMask;
}

template <typename Base, int ElementSize, typename Element>
void CompactArray<Base, ElementSize, Element>::Set(size_t n, Base val) {
  int offset = ElementSize * (n % BaseSize);
  data[n / BaseSize] &= ~(ElemMask << offset);
  data[n / BaseSize] |= val << offset;
}

template <typename Base, int ElementSize, typename Element>
Element CompactArray<Base, ElementSize, Element>::operator[](size_t n) const {
  return Get(n);
}

template <typename Base, int ElementSize, typename Element>
void CompactArray<Base, ElementSize, Element>::Extract(size_t from, size_t length, Element * array) const
{
  for (size_t i = 0; i < length; ++i)
    array[i] = Get(from + i);
}

template <typename Base, int ElementSize, typename Element>
void CompactArray<Base, ElementSize, Element>::Extract(size_t from, size_t length, SArray<Element> & array) const
{
  Extract(from, length, array.data);
}

#endif // _UTIL_COMPACT_HH
