#ifndef _UTIL_COMPACT_CUDA_HH
#define _UTIL_COMPACT_CUDA_HH

#include <stdint.h>

#include "config.hh"
#include "util/siblings.hh"
#include "util/misc.hh"
#include "util/compact_base.hh"
#include "util/compact.hh"

template <typename Base, int ElementSize, typename Element>
struct KernelCompactArray {
  typedef CompactArraySiblings<Base, ElementSize, Element> SiblingTypes;

  KernelArray<Base> data;

  const static size_t BaseSize = sizeof(Base) * 8 / ElementSize;
  const static Base ElemMask = (1ULL << ElementSize) - 1;

  __device__ Element Get(size_t n) const;
#if 0
// RACE
  __device__ void Set(size_t n, Base el);
#endif
  __device__ Element operator[](size_t n) const;

  __device__ void Extract(size_t from, size_t length, Element * array) const;
  __device__ void Extract(size_t from, size_t length, KernelArray<Element> & array) const;
};

template <typename Base, int ElementSize, typename Element>
Element KernelCompactArray<Base, ElementSize, Element>::Get(size_t n) const {
  return (data[n / BaseSize] >> (ElementSize * (n % BaseSize))) & ElemMask;
}

#if 0
// RACE
template <typename Base, int ElementSize, typename Element>
void KernelCompactArray<Base, ElementSize, Element>::Set(size_t n, Base val) {
  int offset = ElementSize * (n % BaseSize);
  data[n / BaseSize] &= ~(ElemMask << offset);
  data[n / BaseSize] |= val << offset;
}
#endif

template <typename Base, int ElementSize, typename Element>
Element KernelCompactArray<Base, ElementSize, Element>::operator[](size_t n) const {
  return Get(n);
}

template <typename Base, int ElementSize, typename Element>
void KernelCompactArray<Base, ElementSize, Element>::Extract(size_t from, size_t length, Element * array) const
{
  for (size_t i = 0; i < length; ++i)
    array[i] = Get(from + i);
}

template <typename Base, int ElementSize, typename Element>
void KernelCompactArray<Base, ElementSize, Element>::Extract(size_t from, size_t length, KernelArray<Element> & array) const
{
  Extract(from, length, array.data);
}

template <typename Base, int ElementSize, typename Element>
class CudaCompactArray {
 public:
  typedef CompactArraySiblings<Base, ElementSize, Element> SiblingTypes;
  typedef typename SiblingTypes::Host HostCA;
  typedef typename SiblingTypes::Kernel KernelCA;

  const static size_t BaseSize = sizeof(Base) * 8 / ElementSize;
  const static Base ElemMask = (1ULL << ElementSize) - 1;

  CudaCompactArray() : size(0) {}
  CudaCompactArray(size_t s);
  CudaCompactArray(const CudaArray<Element> & array);
  CudaCompactArray(const HostCA & ca);

  KernelCA ForKernel() const { KernelCA kc; kc.data = data.ForKernel(); return kc; }

  Element Get(size_t n) const;
  void Set(size_t n, Base el);
  Element operator[](size_t n) const;

  size_t size;
  CudaArray<Base> data;
};

template <typename Base, int ElementSize, typename Element>
Element CudaCompactArray<Base, ElementSize, Element>::Get(size_t n) const {
  return (data[n / BaseSize] >> (ElementSize * (n % BaseSize))) & ElemMask;
}

template <typename Base, int ElementSize, typename Element>
void CudaCompactArray<Base, ElementSize, Element>::Set(size_t n, Base val) {
  int offset = ElementSize * (n % BaseSize);
  Base t = data[n / BaseSize];
  t &= ~(ElemMask << offset);
  t |= val << offset;
  data.Set(n, t);
}

template <typename Base, int ElementSize, typename Element>
Element CudaCompactArray<Base, ElementSize, Element>::operator[](size_t n) const {
  return Get(n);
}

template <typename Base, int ElementSize, typename Element>
CudaCompactArray<Base, ElementSize, Element>::CudaCompactArray(size_t s) :
  size(s), data(ceildiv(size, BaseSize))
{
}

template <typename Base, int ElementSize, typename Element>
CudaCompactArray<Base, ElementSize, Element>::CudaCompactArray(const CudaArray<Element> & array) :
  size(array.size), data(ceildiv(size, BaseSize))
{
  for (size_t i = 0; i < size; ++i)
    Set(i, array[i]);
}

template <typename Base, int ElementSize, typename Element>
CudaCompactArray<Base, ElementSize, Element>::CudaCompactArray(const HostCA & carray) :
  size(carray.size), data(carray.data)
{
}

#endif // _UTIL_COMPACT_CUDA_HH
