#ifndef _DNA_HH
#define _DNA_HH 1

#include <stdint.h>
#include <cstdio>

#include <string>
#include <iostream>
#include <iterator>

#include <boost/serialization/serialization.hpp>

#include "util/array.hh"
#include "util/compact.hh"

typedef uint8_t dna_t;
typedef SArray<dna_t> DnaString;
typedef Array2D<dna_t, 32> DnaStrings;

typedef CompactArray<unsigned long, 2, dna_t> DnaCompact;

template <typename Index>
struct Mapping {
  typename Index::size_type from;
  int id;
  uint errors;
  uint count;
  bool rc;

  Mapping() :
    errors(9999),
    count(0),
    rc(false)
  {}
};

// Proxy to a single entry in Reads
template <typename Index>
struct Read {
  std::string & name;
  DnaStrings::Row data;
  Mapping<Index> & mapping;

  Read(std::string & n, DnaStrings::Row d, Mapping<Index> & m) : name(n), data(d), mapping(m) {}
  Read & operator=(const Read & other) {
    if (this != &other) {
      name = other.name;
      data = other.data;
      mapping = other.mapping;
    }

    return *this;
  }
};

namespace dna {

const dna_t A = 0;
const dna_t C = 1;
const dna_t G = 2;
const dna_t T = 3;
const dna_t N = 4;

const dna_t NOT_DNA = 32;

namespace {
  const dna_t x = NOT_DNA;
  static const dna_t mapper[256] = {
      x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, // 32
      x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, // 64
//       a, b, c, d, e, f, g, h, i, j, k, l, m, n, o, p, q, r, s, t, u, v, w, x, y, z
      x, A, N, C, N, x, x, G, N, x, x, N, x, N, N, x, x, x, N, N, T, N, N, N, N, N, x, x, x, x, x, x, // 96
      x, A, N, C, N, x, x, G, N, x, x, N, x, N, N, x, x, x, N, N, T, N, N, N, N, N, x, x, x, x, x, x, // 128
      x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, // 160
      x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, // 192
      x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, // 224
      x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, x, // 256
  };

  static const dna_t complement[] = {T, G, C, A, N};

  static const char unmapper[] = {
      'A', 'C', 'G', 'T', 'N'
  };
}

dna_t Complement(dna_t d) {
  return complement[d];
}

// from chars to our 2bit alphabet
dna_t Map(uint8_t c) {
    return mapper[c];
}

DnaString Map(const std::string & s) {
  DnaString d(s.size());
  for (size_t i = 0 ; i < s.size() ; ++i)
    d[i] = Map(s[i]);

  return d;
}

char Unmap(dna_t dna) {
    return unmapper[dna];
}

std::string Unmap(const DnaString & d) {
  std::string s(d.size, 'x');
  for (DnaString::size_type i = 0 ; i < d.size ; ++i)
    s[i] = Unmap(d[i]);
  return s;
}

template <typename A>
std::string Unmap(const A & d) {
  std::string s(d.size(), 'x');
  for (DnaString::size_type i = 0 ; i < d.size() ; ++i)
    s[i] = Unmap(d[i]);
  return s;
}

DnaString ReadDna(std::istream & is) {
  char buf[1 << 20];
  int r;

  std::streampos pos = is.tellg();
  size_t count = 0;

  while ((r = is.read(buf, sizeof(buf)).gcount()) > 0) {
    for (int i = 0; i < r ; ++i) {
      uint8_t c = Map(buf[i]);
      if (c != NOT_DNA)
        ++count;
    }
  }

  is.clear();
  is.seekg(pos);

  DnaString dna(count);
  count = 0;

  while ((r = is.read(buf, sizeof(buf)).gcount()) > 0) {
    for (int i = 0; i < r ; ++i) {
      uint8_t c = Map(buf[i]);
      if (c != NOT_DNA)
        dna[count++] = c;
    }
  }

  return dna;
}

DnaString & Replace(DnaString & d, dna_t from, dna_t to) {
  for (size_t i = 0 ; i < d.size ; ++i)
    if (d[i] == from)
      d[i] = to;
  return d;
}

DnaString Strip(const DnaString & d, dna_t what) {
  size_t j = 0;
  for (size_t i = 0 ; i < d.size ; ++i)
    if (d[i] == what)
      ++j;
  DnaString ds(d.size - j);
  j = 0;
  for (size_t i = 0 ; i < d.size ; ++i)
    if (d[i] != what)
      ds[j++] = d[i];
  return ds;
}

DnaString & Randomize(DnaString & d) {
  for (size_t i = 0 ; i < d.size ; ++i)
    d[i] = rand() % 4;
  return d;
}

DnaString & RandomizeN(DnaString & d) {
  for (size_t i = 0 ; i < d.size ; ++i)
    if (d[i] == N)
      d[i] = rand() % 4;
  return d;
}

bool HasN(DnaString & d) {
  for (size_t i = 0 ; i < d.size ; ++i)
    if (d[i] == N)
      return true;
  return false;
}

template <typename A>
bool HasN(A d) {
  for (size_t i = 0; i < d.size(); ++i)
    if (d[i] == N)
      return true;
  return false;
}

DnaString & ReverseComplement(DnaString & d) {
  size_t l = 0;
  size_t r = d.size - 1;
  for ( ; l < r ; ++l, --r) {
    dna_t t = d[l];
    d[l] = Complement(d[r]);
    d[r] = Complement(t);
  }
  if (l == r)
    d[l] = Complement(d[l]);

  return d;
}

template <typename A1, typename A2>
void ReverseComplement(A1 d1, A2 d2) {
  for (uint i = 0; i < d2.size(); ++i)
      d2[i] = Complement(d1[d1.size() - i - 1]);
}

struct DnaRead {
  std::string name;
  DnaString data[2];
};

class FastqReader {
 public:
  FastqReader(std::istream & is_) :
    is(is_),
    pos(0), to_read(0)
  {
    tmp.reserve(512);
  }

  // used by mapper
  bool Next(DnaRead & read) {
    read.name = get_line();
    if (read.name.size() > 0)
      read.name = read.name.substr(1);
    read.data[0] = dna::Map(get_line());
    skip_line();
    skip_line();

    return read.data[0].size > 0;
  }

  // used mapper_cuda
  template <typename read_type>
  bool Next(read_type read) {
    read.name = get_line();
    if (read.name.size() > 0)
      read.name = read.name.substr(1);
    // TODO handle reads longer than expected
    bool notempty = get_dna(read.data);
    skip_line();
    // TODO save  qualities
    skip_line();

    return notempty;
  }

  std::string get_line() {
    tmp.clear();
    for (;pos < to_read || refill(); ++pos) {
      char c = buf[pos];
      if (c == '\n') {
        ++pos;
        break;
      }
      tmp.push_back(c);
    }
    return tmp;
  }

  template <typename Array>
  bool get_dna(Array dna_data) {
    uint i = 0;
    for (; pos < to_read || refill(); ++pos) {
      char c = buf[pos];
      if (c == '\n') {
        ++pos;
        break;
      }
      dna_data[i++] = dna::Map(c);
    }
    return i > 0;
  }

  void skip_line() {
    for (;pos < to_read || refill(); ++pos) {
      if (buf[pos] == '\n') {
        ++pos;
        break;
      }
    }
  }

 private:
  bool refill() {
    pos = 0;
    return (to_read = is.read(buf, sizeof(buf)).gcount()) > 0;
  }

  char buf[1 << 20];
  std::string tmp;
  std::istream & is;
  int pos;
  int to_read;
};


// Info about a single chromosome
class SequenceMeta {
public:
  SequenceMeta() : label(""), size(0) {} 
  SequenceMeta(std::string l, size_t s) :
    label(l), size(s) {}

  std::string label;
  size_t size;

  friend class boost::serialization::access;
  template <typename Archive>
  void serialize(Archive & ar, const unsigned int version);
};

template <typename Archive>
void SequenceMeta::serialize(Archive & ar, const unsigned int version) {
  ar & label;
  ar & size;
}


class FastaReader {
 public:
  FastaReader(std::istream & is_) :
    is(is_),
    pos(0), to_read(0), size(0),
    eof(false), meta_read(false)
  {
  }

  void read_meta()
  {
    std::cerr << "Reading meta data..." << std::flush;
    size = 0;
    size_t seq_size = 0;

    std::streampos pos = is.tellg();
    tmp.reserve(512);

    std::string last_label = "", line;

    while(true)
    {
      line = get_line();
      if(eof)
      {
        if(last_label.size() > 0)
          meta.push_back(SequenceMeta(last_label, seq_size));
        size += seq_size;
        break;
      }

      if(line.size() > 0 && line[0] == '>')
      {
        if(last_label.size() > 0)
          meta.push_back(SequenceMeta(last_label, seq_size));

        last_label = parse_label(line);

        size += seq_size;
        seq_size = 0;
      }
      else
      {
        for(size_t i = 0; i < line.size(); i++)
        {
          uint8_t c = Map(line[i]);
          if (c != NOT_DNA)
            ++seq_size;
        }
      }
    }

    is.clear();
    is.seekg(pos);

    std::cerr << "done\n";

    for(size_t i = 0; i < meta.size(); i++)
      std::cerr << "Sequence: " << meta[i].label << " Size: " << meta[i].size << std::endl;
    std::cerr << "Total size: " << size << std::endl;
    meta_read = true;
  }

  void read_dna(DnaString & dna)
  {
    if(!meta_read)
      read_meta();

    std::cerr << "Reading DNA..." << std::flush;
    dna.Resize(size);

    std::streampos pos = is.tellg();
    eof = false;

    size_t count = 0;
    while(true)
    {
      std::string line = get_line();

      if(eof)
        break;

      if(line.size() > 0 && line[0] != '>')
      {
        for(size_t i = 0; i < line.size(); i++)
        {
          uint8_t c = Map(line[i]);
          if (c != NOT_DNA)
            dna[count++] = c;
        }
      }
    }

    is.clear();
    is.seekg(pos);

    std::cerr << "done\n";
  }

  size_t get_size() { return size; }
  void get_meta(SArray<SequenceMeta> & meta_out)
  {
    if(!meta_read)
      read_meta();

    meta_out = SArray<SequenceMeta>(meta);
  }

  std::string get_line() {
    tmp.clear();
    for (;pos < to_read || refill(); ++pos) {
      char c = buf[pos];
      if (c == '\n') {
        ++pos;
        break;
      }
      tmp.push_back(c);
    }
    return tmp;
  }

 private:
  bool refill() {
    pos = 0;
    eof = (to_read = is.read(buf, sizeof(buf)).gcount()) == 0;
    return !eof;
  }

  std::string parse_label(std::string & line)
  {
    size_t pos = line.find(" ");
    if(pos == std::string::npos)
      pos = line.size();
    return line.substr(1, pos - 1);
  }

  char buf[1 << 20];
  std::string tmp;
  std::istream & is;
  int pos;
  int to_read;
  size_t size;
  bool eof, meta_read;
  std::vector<SequenceMeta> meta;
};

} // namespace dna

#endif
